import sqlite3
import unittest

from MiscUtils.DBPool import DBPool


class TestDBPool(unittest.TestCase):

    def testDbPool(self):
        pool = DBPool(sqlite3, 10, database=':memory:')
        for _count in range(15):
            con = pool.connection()
            cursor = con.cursor()
            cursor.execute("select 1 union select 2 union select 3 order by 1")
            rows = cursor.fetchall()
            self.assertEqual(rows, [(1,), (2,), (3,)])
            con.close()



import unittest

from io import BytesIO

from SessionStore import SessionStore

from .Application import Application
from .Session import Session


class TestSessionStore(unittest.TestCase):

    _storeClass = SessionStore

    _app = Application()

    def setUp(self):
        Session._lastExpired = None
        self._store = self._storeClass(self._app)

    def testApplication(self):
        self.assertEqual(self._store.application(), self._app)

    def testEncodeDecode(self):
        session = Session()
        f = BytesIO()
        self._store.encoder()(session, f)
        output = f.getvalue()
        f.close()
        self.assertTrue(isinstance(output, bytes))
        f = BytesIO(output)
        output = self._store.decoder()(f)
        f.close()
        self.assertTrue(type(output) is type(session))
        self.assertEqual(output._data, session._data)

    def testSetEncoderDecoder(self):
        def encoder(obj, f):
            return f.write(repr(obj).encode('ascii'))

        def decoder(f):
            return eval(f.read().decode('ascii'))

        self._store.setEncoderDecoder(encoder, decoder)
        self.assertEqual(self._store.encoder(), encoder)
        self.assertEqual(self._store.decoder(), decoder)



import os
import sys

from MiscUtils.PropertiesObject import PropertiesObject


class PlugInError(Exception):
    """Plug-in error."""


class PlugIn:
    """Template for Webware Plug-ins.

    A plug-in is a software component that is loaded by Webware in order to
    provide additional Webware functionality without necessarily having to
    modify Webware's source. The most infamous plug-in is PSP (Python Server
    Pages) which ships with Webware.

    Plug-ins often provide additional servlet factories, servlet subclasses,
    examples and documentation. Ultimately, it is the plug-in author's choice
    as to what to provide and in what manner.

    Instances of this class represent plug-ins which are ultimately Python
    packages.

    A plug-in must also be a Webware component which means that it will have
    a Properties.py file advertising its name, version, requirements, etc.
    You can ask a plug-in for its properties().

    The plug-in/package must have an __init__.py which must contain the
    following function::

        def installInWebware(application):
            ...

    This function is invoked to take whatever actions are needed to plug the
    new component into Webware. See PSP for an example.

    If you ask an Application for its plugIns(), you will get a list of
    instances of this class.

    The path of the plug-in is added to sys.path, if it's not already there.
    This is convenient, but we may need a more sophisticated solution in the
    future to avoid name collisions between plug-ins.

    Note that this class is hardly ever subclassed. The software in the
    plug-in package is what provides new functionality and there is currently
    no way to tell the Application to use custom subclasses of this class on a
    case-by-case basis (and so far there is currently no need).

    Instructions for invoking::

        # 'self' is typically Application. It gets passed to installInWebware()
        p = PlugIn(self, 'Foo', '../Foo')
        willNotLoadReason = plugIn.load()
        if willNotLoadReason:
            print(f'Plug-in {path} cannot be loaded because:')
            print(willNotLoadReason)
            return None
        p.install()
        # Note that load() and install() could raise exceptions.
        # You should expect this.
    """

    # region Init, load and install

    def __init__(self, application, name, module):
        """Initializes the plug-in with basic information.

        This lightweight constructor does not access the file system.
        """
        self._app = application
        self._name = name
        self._module = module
        self._path = module.__path__[0]
        try:
            self._builtin = module.__package__ == f'webware.{name}'
        except AttributeError:
            self._builtin = False
        self._dir = os.path.dirname(self._path)
        self._cacheDir = os.path.join(self._app._cacheDir, self._name)
        self._examplePages = self._examplePagesContext = None

    def load(self, verbose=True):
        """Loads the plug-in into memory, but does not yet install it.

        Will return None on success, otherwise a message (string) that says
        why the plug-in could not be loaded.
        """
        if verbose:
            print(f'Loading plug-in: {self._name} at {self._path}')

        # Grab the Properties.py
        self._properties = PropertiesObject(
            self.serverSidePath('Properties.py'))

        if self._builtin and 'version' not in self._properties:
            self._properties['version'] = self._app._webwareVersion
            self._properties['versionString'] = self._app._webwareVersionString

        if not self._properties['willRun']:
            return self._properties['willNotRunReason']

        # Update sys.path
        if self._dir not in sys.path:
            sys.path.append(self._dir)

        # Inspect it and verify some required conventions
        if not hasattr(self._module, 'installInWebware'):
            raise PlugInError(
                f"Plug-in '{self._name!r}' in {self._dir!r}"
                " has no installInWebware() function.")

        # Give the module a pointer back to us
        setattr(self._module, 'plugIn', self)

        # Make a subdirectory for it in the Cache directory:
        if not os.path.exists(self._cacheDir):
            os.mkdir(self._cacheDir)

        self.setUpExamplePages()

    def setUpExamplePages(self):
        """Add a context for the examples."""
        if self._app.hasContext('Examples'):
            config = self._properties.get('webwareConfig', {})
            self._examplePages = config.get('examplePages') or None
            if self.hasExamplePages():
                examplesPath = self.serverSidePath('Examples')
                if not os.path.exists(examplesPath):
                    raise PlugInError(
                        f'Plug-in {self._name!r} says it has example pages, '
                        'but there is no Examples/ subdir.')
                if os.path.exists(os.path.join(examplesPath, '__init__.py')):
                    ctxName = self._name + '/Examples'
                    if not self._app.hasContext(ctxName):
                        self._app.addContext(ctxName, examplesPath)
                    self._examplePagesContext = ctxName
                else:
                    raise PlugInError(
                        'Cannot create Examples context for'
                        f' plug-in {self._name!r} (no __init__.py found).')

    def examplePages(self):
        return self._examplePages

    def hasExamplePages(self):
        return self._examplePages is not None

    def examplePagesContext(self):
        return self._examplePagesContext

    def install(self):
        """Install plug-in by invoking its installInWebware() function."""
        self._module.installInWebware(self._app)

    # endregion Init, load and install

    # region Access

    def name(self):
        """Return the name of the plug-in. Example: 'Foo'"""
        return self._name

    def directory(self):
        """Return the directory in which the plug-in resides. Example: '..'"""
        return self._dir

    def path(self):
        """Return the full path of the plug-in. Example: '../Foo'"""
        return self._path

    def serverSidePath(self, path=None):
        if path:
            return os.path.normpath(os.path.join(self._path, path))
        return self._path

    def module(self):
        """Return the Python module object of the plug-in."""
        return self._module

    def properties(self):
        """Return the properties.

        This is a dictionary-like object, of the plug-in which comes
        from its Properties.py file. See MiscUtils.PropertiesObject.py.
        """
        return self._properties

    # endregion Access




from SidebarPage import SidebarPage


class FieldStorage(SidebarPage):
    """Test the FieldStorage class.

    Webware uses a modified FieldStorage class that parses GET parameters
    even in a POST request. However, the parameter names must be different.
    A GET parameters with the same name as a POST parameter is ignored,
    these values are not appended. In other words, POST parameters always
    override GET parameters with the same name.
    """

    def cornerTitle(self):
        return 'Testing'

    def writeContent(self):
        request = self.request()
        method = request.method()
        fields = request.fields()
        wr = self.writeln
        if method == 'GET' and not fields:
            getFields = [
                ('getfield1', 'getvalue1'),
                ('getfield2', 'getvalue21'), ('getfield2', 'getvalue22'),
                ('dualfield1', 'getdual1'),
                ('dualfield2', 'getdual21'), ('dualfield2', 'getdual22'),
                ('getempty', '')
            ]
            postFields = [
                ('postfield1', 'postvalue1'),
                ('postfield2', 'postvalue21'), ('postfield2', 'postvalue22'),
                ('dualfield1', 'postdual1'),
                ('dualfield2', 'postdual21'), ('dualfield2', 'postdual22'),
                ('postempty', '')
            ]
            wr('<p>The Webware FieldStorage class can be tested here.</p>')
            wr('<form action="FieldStorage?{}" method="POST">'.format(
                '&amp;'.join('{}={}'.format(*field) for field in getFields)))
            wr('<p>Please press the button to run the test:')
            for field in postFields:
                wr('<input type="hidden" name="{}" value="{}">'.format(*field))
            wr('<input type="submit" name="testbutton" value="Submit">')
            wr('</p></form>')
        else:
            errors = []
            error = errors.append
            if method != 'POST':
                error(f'The method is {method} instead of POST')
            if len(fields) != 9:
                error(f'We got {len(fields)} instead of 9 fields')
            if not request.hasField('testbutton'):
                error('We did not get the submit button')
            elif request.field('testbutton') != 'Submit':
                error('The submit button field got a wrong value')
            if not request.hasField('getempty'):
                error('We did not get the empty GET parameter')
            elif request.field('getempty') != '':
                error('The empty GET field got a non-empty value')
            if not request.hasField('postempty'):
                error('We did not get the empty POST parameter')
            elif request.field('postempty') != '':
                error('The empty POST field got a non-empty value')
            if not request.hasField('getfield1'):
                error('The first GET parameter has not been passed')
            elif request.field('getfield1') != 'getvalue1':
                error('The first GET field got a wrong value')
            if not request.hasField('postfield1'):
                error('The first POST parameter has not been passed')
            elif request.field('postfield1') != 'postvalue1':
                error('The first POST field got a wrong value')
            if not request.hasField('getfield2'):
                error('The second GET parameter has not been passed')
            elif request.field('getfield2') != ['getvalue21', 'getvalue22']:
                error('The second GET field got a wrong value')
            if not request.hasField('postfield2'):
                error('The second POST parameter has not been passed')
            elif request.field('postfield2') != ['postvalue21', 'postvalue22']:
                error('The second POST field got a wrong value')
            if not request.hasField('dualfield1'):
                error('The first dual parameter has not been passed')
            elif request.field('dualfield1') == 'getdual1':
                error('The first dual field was not overridden via POST')
            elif request.field('dualfield1') in (
                    ['getdual1', 'postdual1'], ['postdual1', 'getdual1']):
                error('The first dual field'
                      ' was extended instead of overridden via POST')
            elif request.field('dualfield1') != 'postdual1':
                error('The first dual field got a wrong value')
            if not request.hasField('dualfield2'):
                error('The second dual parameter has not been passed')
            elif request.field('dualfield2') == ['getdual21', 'getdual22']:
                error('The second dual field was not overridden via POST')
            elif request.field('dualfield2') in (
                    ['getdual21', 'getdual22', 'postdual21', 'postdual22'],
                    ['postdual21', 'postdual22', 'getdual21', 'getdual22']):
                error('The second dual field'
                      ' was extended instead of overridden via POST')
            elif request.field('dualfield2') != ['postdual21', 'postdual22']:
                error('The second dual field got a wrong value')
            if errors:
                wr('<p>FieldStorage does <em>not</em> work as expected:</p>')
                wr('<ul>')
                for error in errors:
                    wr(f'<li>{error}.</li>')
                wr('</ul>')
            else:
                wr('<p>Everything ok, FieldStorage works as expected.</p>')
            wr('<p><a href="./">Back to the test cases overview.</a></p>')



"""A general session store."""

from pickle import load, dump, HIGHEST_PROTOCOL as maxPickleProtocol
from warnings import warn

from time import time

from MiscUtils import AbstractError


def dumpWithHighestProtocol(obj, f):
    """Same as pickle.dump, but by default with the highest protocol."""
    return dump(obj, f, maxPickleProtocol)


class SessionStore:
    """A general session store.

    SessionStores are dictionary-like objects used by Application to
    store session state. This class is abstract and it's up to the
    concrete subclass to implement several key methods that determine
    how sessions are stored (such as in memory, on disk or in a
    database). We assume that session keys are always strings.

    Subclasses often encode sessions for storage somewhere. In light
    of that, this class also defines methods encoder(), decoder() and
    setEncoderDecoder(). The encoder and decoder default to the load()
    and dump() functions of the pickle module. However, using the
    setEncoderDecoder() method, you can use the functions from marshal
    (if appropriate) or your own encoding scheme. Subclasses should use
    encoder() and decoder() (and not pickle.load() and pickle.dump()).

    Subclasses may rely on the attribute self._app to point to the
    application.

    Subclasses should be named SessionFooStore since Application
    expects "Foo" to appear for the "SessionStore" setting and
    automatically prepends Session and appends Store. Currently, you
    will also need to add another import statement in Application.py.
    Search for SessionStore and you'll find the place.

    TO DO

      * Should there be a check-in/check-out strategy for sessions to
        prevent concurrent requests on the same session? If so, that can
        probably be done at this level (as opposed to pushing the burden
        on various subclasses).
    """

    # region Init

    def __init__(self, app):
        """Initialize the session store.

        Subclasses must invoke super.
        """
        self._app = app
        self._alwaysSave = app._alwaysSaveSessions
        self._retain = app._retainSessions
        self._encoder = dumpWithHighestProtocol
        self._decoder = load

    # endregion Init

    # region Access

    def application(self):
        """Return the application owning the session store."""
        return self._app

    # endregion Access

    # region Dictionary-style access

    def __len__(self):
        """Return the number of sessions in the store.

        Subclasses must implement this method.
        """
        raise AbstractError(self.__class__)

    def __getitem__(self, key):
        """Get a session item from the store.

        Subclasses must implement this method.
        """
        raise AbstractError(self.__class__)

    def __setitem__(self, key, value):
        """Set a session item, saving it to the store.

        Subclasses must implement this method.
        """
        raise AbstractError(self.__class__)

    def __delitem__(self, key):
        """Delete a session item from the store.

        Subclasses are responsible for expiring the session as well.
        Something along the lines of:
            session = self[key]
            if not session.isExpired():
                session.expiring()
        """
        raise AbstractError(self.__class__)

    def __contains__(self, key):
        """Check whether the session store has a given key.

        Subclasses must implement this method.
        """
        raise AbstractError(self.__class__)

    def __iter__(self):
        """Return an iterator over the stored session keys.

        Subclasses must implement this method.
        """
        raise AbstractError(self.__class__)

    def has_key(self, key):
        """Check whether the session store has a given key."""
        warn("has_key is deprecated, please us 'in' instead.",
             DeprecationWarning, stacklevel=2)
        return key in self

    def keys(self):
        """Return a list with the keys of all the stored sessions.

        Subclasses must implement this method.
        """
        raise AbstractError(self.__class__)

    def iterkeys(self):
        """Return an iterator over the stored session keys."""
        return iter(self)

    def clear(self):
        """Clear the session store, removing all of its items.

        Subclasses must implement this method.
        """
        raise AbstractError(self.__class__)

    def setdefault(self, key, default=None):
        """Return value if key available, else default (also setting it).

        Subclasses must implement this method.
        """
        raise AbstractError(self.__class__)

    def pop(self, key, default=None):
        """Return value if key available, else default (also remove key).

        Subclasses must implement this method.
        """
        raise AbstractError(self.__class__)

    # endregion Dictionary-style access

    # region Application support

    def storeSession(self, session):
        """Save potentially changed session in the store.

        Used at the end of transactions.

        Subclasses must implement this method.
        """
        raise AbstractError(self.__class__)

    def storeAllSessions(self):
        """Permanently save all sessions in the store.

        Used when the application server is shut down.

        Subclasses must implement this method.
        """
        raise AbstractError(self.__class__)

    def cleanStaleSessions(self, _task=None):
        """Clean stale sessions.

        Called by the Application to tell this store to clean out all
        sessions that have exceeded their lifetime.
        """
        curTime = time()
        keys = []
        for key in self.keys():
            try:
                session = self[key]
            except KeyError:
                pass  # session was already deleted by some other thread
            else:
                try:
                    timeout = session.timeout()
                    if timeout is not None and (
                            timeout == 0 or
                            curTime >= session.lastAccessTime() + timeout):
                        keys.append(key)
                except AttributeError as e:
                    raise ValueError(
                        f'Not a Session object: {session!r}') from e
        for key in keys:
            try:
                del self[key]
            except KeyError:
                pass  # already deleted by some other thread

    # endregion Application support

    # region Convenience methods

    def get(self, key, default=None):
        """Return value if key available, else return the default."""
        try:
            return self[key]
        except KeyError:
            return default

    def items(self):
        """Return a list with the (key, value) pairs for all sessions."""
        return list(self.iteritems())

    def values(self):
        """Return a list with the values of all stored sessions."""
        return list(self.itervalues())

    def iteritems(self):
        """Return an iterator over the (key, value) pairs for all sessions."""
        for key in self.keys():
            # since the size of the store might change during iteration,
            # we iterate over the list of keys, not the store itself
            try:
                yield key, self[key]
            except KeyError:
                # since we aren't using a lock here, some keys
                # could be already deleted again during this loop
                pass

    def itervalues(self):
        """Return an iterator over the stored values of all sessions."""
        for key in self.keys():
            try:
                yield self[key]
            except KeyError:
                pass

    # endregion Convenience methods

    # region Encoder/decoder

    def encoder(self):
        """Return the value serializer for the store."""
        return self._encoder

    def decoder(self):
        """Return the value deserializer for the store."""
        return self._decoder

    def setEncoderDecoder(self, encoder, decoder):
        """Set the serializer and deserializer for the store."""
        self._encoder = encoder
        self._decoder = decoder

    # endregion Encoder/decoder

    # region As a string

    def __repr__(self):
        """Return string representation of the store like a dictionary."""
        return repr(dict(self.items()))

    # endregion As a string



"""Dict-RPC servlets."""

import sys
import traceback
from time import time

try:
    from pickle import dumps, PickleError
except ImportError:
    from pickle import dumps, PickleError

try:
    import zlib
except ImportError:
    zlib = None

from RPCServlet import RPCServlet
from MiscUtils.PickleRPC import RequestError, SafeUnpickler


class PickleRPCServlet(RPCServlet, SafeUnpickler):
    """PickleRPCServlet is a base class for Dict-RPC servlets.

    The "Pickle" refers to Python's pickle module. This class is
    similar to XMLRPCServlet. By using Python pickles you get their
    convenience (assuming the client is Pythonic), but lose
    language independence. Some of us don't mind that last one.  ;-)

    Conveniences over XML-RPC include the use of all of the following:

      * Any pickle-able Python type (datetime for example)
      * Python instances (aka objects)
      * None
      * Longs that are outside the 32-bit int boundaries
      * Keyword arguments

    Pickles should also be faster than XML, especially now that
    we support binary pickling and compression.

    To make your own `PickleRPCServlet`, create a subclass and implement a
    method which is then named in `exposedMethods()`::

        from PickleRPCServlet import PickleRPCServlet
        class Math(PickleRPCServlet):
            def multiply(self, x, y):
                return x * y
            def exposedMethods(self):
                return ['multiply']

    To make a PickleRPC call from another Python program, do this::

        from MiscUtils.PickleRPC import Server
        server = Server('http://localhost/Webware/Context/Math')
        print(server.multiply(3, 4))     # 12
        print(server.multiply('-', 10))  # ----------

    If a request error is raised by the server, then
    `MiscUtils.PickleRPC.RequestError` is raised. If an unhandled
    exception is raised by the server, or the server response is
    malformed, then `MiscUtils.PickleRPC.ResponseError` (or one of
    its subclasses) is raised.

    Tip: If you want callers of the RPC servlets to be able to
    introspect what methods are available, then include
    'exposedMethods' in `exposedMethods()`.

    If you wanted the actual response dictionary for some reason::

        print(server._request('multiply', 3, 4))
        # {'value': 12, 'timeReceived': ...}

    In which case, an exception is not purposefully raised if the
    dictionary contains one. Instead, examine the dictionary.

    For the dictionary formats and more information see the docs
    for `MiscUtils.PickleRPC`.
    """

    def respondToPost(self, trans):
        try:
            request = trans.request()
            data = request.rawInput(rewind=1)
            response = dict(timeReceived=trans.request().time())
            try:
                try:
                    encoding = request.environ().get('HTTP_CONTENT_ENCODING')
                    if encoding == 'x-gzip':
                        if zlib is not None:
                            try:
                                rawstring = data.read()
                                req = self.loads(zlib.decompress(rawstring))
                            except zlib.error as e:
                                raise RequestError(
                                    'Cannot uncompress'
                                    ' compressed dict-rpc request') from e
                        else:
                            raise RequestError(
                                'Cannot handle compressed dict-rpc request')
                    elif encoding:
                        raise RequestError(
                            f'Cannot handle Content-Encoding of {encoding}')
                    else:
                        req = self.load(data)
                except PickleError as e:
                    raise RequestError(
                        'Cannot unpickle dict-rpc request.') from e
                if not isinstance(req, dict):
                    raise RequestError(
                        'Expecting a dictionary for dict-rpc requests, '
                        f'but got {type(req)} instead.')
                if req.get('version', 1) != 1:
                    raise RequestError(
                        f"Cannot handle version {req['version']} requests.")
                if req.get('action', 'call') != 'call':
                    raise RequestError(
                        f"Cannot handle the request action, {req['action']}.")
                try:
                    methodName = req['methodName']
                except KeyError:
                    raise RequestError(
                        'Missing method name in request') from None
                args = req.get('args', ())
                if methodName == '__methods__.__getitem__':
                    # support PythonWin autoname completion
                    response['value'] = self.exposedMethods()[args[0]]
                else:
                    response['value'] = self.call(
                        methodName, *args, **req.get('keywords', {}))
            except RequestError as e:
                response['requestError'] = str(e)
                self.sendResponse(trans, response)
                self.handleException(trans)
            except Exception as e:
                response['exception'] = self.resultForException(e, trans)
                self.sendResponse(trans, response)
                self.handleException(trans)
            else:
                self.sendResponse(trans, response)
        except Exception:
            # internal error, report as HTTP server error
            print('PickleRPCServlet internal error')
            print(''.join(traceback.format_exception(*sys.exc_info())))
            trans.response().setStatus(500, 'Server Error')
            self.handleException(trans)

    def sendResponse(self, trans, response):
        """Timestamp the response dict and send it."""
        # Generated a pickle string
        response['timeResponded'] = time()
        if self.useBinaryPickle():
            contentType = 'application/x-python-binary-pickled-dict'
            response = dumps(response, -1)
        else:
            contentType = 'text/x-python-pickled-dict'
            response = dumps(response, 0)

        # Get list of accepted encodings
        try:
            acceptEncoding = trans.request().environ()["HTTP_ACCEPT_ENCODING"]
            if acceptEncoding:
                acceptEncoding = [
                    enc.strip() for enc in acceptEncoding.split(',')]
            else:
                acceptEncoding = []
        except KeyError:
            acceptEncoding = []

        # Compress the output if we are allowed to.
        # We'll avoid compressing short responses and
        # we'll use the fastest possible compression -- level 1.
        if (zlib is not None and 'gzip' in acceptEncoding
                and len(response) > 1000):
            contentEncoding = 'x-gzip'
            response = zlib.compress(response, 1)
        else:
            contentEncoding = None
        self.sendOK(contentType, response, trans, contentEncoding)

    @staticmethod
    def useBinaryPickle():
        """Determine whether binary pickling format shall be used.

        When this returns True, the highest available binary pickling format
        will be used. Override this to return False to use the less-efficient
        text pickling format.
        """
        return True



from Page import Page


class DebugPage(Page):

    def state(self):
        envVars = ('PATH_INFO', 'REQUEST_URI', 'SCRIPT_NAME')
        reqVars = (
            'urlPath', 'previousURLPaths',
            'scriptName', 'servletPath', 'contextName',
            'serverPath', 'serverSidePath', 'serverSideContextPath',
            'extraURLPath')
        req = self.request()
        env = req._environ
        s = []
        for key in envVars:
            value = env.get(key, "* not set *")
            s.append(f"  * env[{key!r}] = {value}")
        for key in reqVars:
            value = getattr(req, key)()
            s.append(f"  * req.{key}() = {value}")
        return '\n'.join(s)

    def writeContent(self):
        self.writeln(f'<h2><code>{self.__class__.__name__}</code></h2>')
        self.writeln(f'<pre>{self.state()}</pre>')



from Page import Page


class IncludeMe(Page):

    def writeHTML(self):
        self.writeln('<h1>Hello from IncludeMe</h1>')
        self.writeln('<p>I like to be included. This is my content.</p>')



"""An abstract request"""

from MiscUtils import AbstractError
from MiscUtils.Funcs import asclocaltime


class Request:
    """The abstract request class.

    Request is a base class that offers the following:

      * A time stamp (indicating when the request was made)
      * An input stream
      * Remote request information (address, name)
      * Local host information (address, name, port)
      * A security indicator

    Request is an abstract class; developers typically use HTTPRequest.
    """

    # region Init

    def __init__(self):
        """Initialize the request.

        Subclasses are responsible for invoking super
        and initializing self._time.
        """
        self._transaction = None

    # endregion Init

    # region Access

    def time(self):
        return self._time  # pylint: disable=no-member

    def timeStamp(self):
        """Return time() as human readable string for logging and debugging."""
        return asclocaltime(self.time())

    # endregion Access

    # region Input

    def input(self):
        """Return a file-style object that the contents can be read from."""
        # This is bogus. Disregard for now.

    # endregion Input

    # region Remote info

    def remoteAddress(self):
        """Get the remote address.

        Returns a string containing the Internet Protocol (IP) address
        of the client that sent the request.
        """
        raise AbstractError(self.__class__)

    def remoteName(self):
        """Get the remote name.

        Returns the fully qualified name of the client that sent the request,
        or the IP address of the client if the name cannot be determined.
        """
        raise AbstractError(self.__class__)

    # endregion Remote info

    # region Local info

    def localAddress(self):
        """Get local address.

        Returns a string containing the Internet Protocol (IP) address
        of the local host (e.g., the server) that received the request.
        """
        raise AbstractError(self.__class__)

    @staticmethod
    def localName():
        """Get local name.

        Returns the fully qualified name of the local host (e.g., the server)
        that received the request.
        """
        return 'localhost'

    def localPort(self):
        """Get local port.

        Returns the port of the local host (e.g., the server)
        that received the request.
        """
        raise AbstractError(self.__class__)

    # endregion Local info

    # region Security

    @staticmethod
    def isSecure():
        """Check whether this is a secure channel.

        Returns true if request was made using a secure channel,
        such as HTTPS. This currently always returns false,
        since secure channels are not yet supported.
        """
        return False

    # endregion Security

    # region Transactions

    def responseClass(self):
        """Get the corresponding response class."""
        raise AbstractError(self.__class__)

    def setTransaction(self, trans):
        """Set a transaction container."""
        self._transaction = trans

    def transaction(self):
        """Get the transaction container."""
        return self._transaction

    # endregion Transactions

    # region Cleanup

    def clearTransaction(self):
        del self._transaction

    # endregion Cleanup

    # region Exception reports

    _exceptionReportAttrNames = []

    def writeExceptionReport(self, handler):
        handler.writeTitle(self.__class__.__name__)
        handler.writeAttrs(self, self._exceptionReportAttrNames)

    # endregion Exception reports



"""The abstract UserManager class."""

from MiscUtils import AbstractError, NoDefault
from .User import User


class UserManager:
    """The base class for all user manager classes.

    A UserManager manages a set of users including authentication,
    indexing and persistence. Keep in mind that UserManager is abstract;
    you must always use a concrete subclasses like UserManagerToFile
    (but please read the rest of this docstring).

    You can create a user through the manager (preferred)::

        user = manager.createUser(name, password)

    Or directly through the user class::

        user = RoleUser(manager, name, password)
        manager.addUser(user)

    The manager tracks users by whether or not they are "active"
    (e.g., logged in) and indexes them by:

      * user serial number
      * external user id
      * user name

    These methods provide access to the users by these keys::

        def userForSerialNum(self, serialNum, default=NoDefault)
        def userForExternalId(self, extId, default=NoDefault)
        def userForName(self, name, default=NoDefault)

    UserManager provides convenient methods for iterating through the
    various users. Each method returns an object that can be used in a
    for loop and asked for its len()::

        def users(self)
        def activeUsers(self)
        def inactiveUsers(self)

    You can authenticate a user by passing the user object and attempted
    password to login(). If the authentication is successful, then login()
    returns the User, otherwise it returns None::

        user = mgr.userForExternalId(externalId)
        if mgr.login(user, password):
            self.doSomething()

    As a convenience, you can authenticate by passing the serialNum,
    externalId or name of the user::

        def loginSerialNum(self, serialNum, password):
        def loginExternalId(self, externalId, password):
        def loginName(self, userName, password):

    The user will automatically log out after a period of inactivity
    (see below), or you can make it happen with::

        def logout(self, user):

    There are three user states that are important to the manager:

      * modified
      * cached
      * authenticated or "active"

    A modified user is one whose data has changed and eventually requires
    storage to a persistent location. A cached user is a user whose data
    resides in memory (regardless of the other states). An active user
    has been authenticated (e.g., their username and password were checked)
    and has not yet logged out or timed out.

    The manager keeps three timeouts, expressed in minutes, to:

      * save modified users after a period of time following the first
        unsaved modification
      * push users out of memory after a period of inactivity
      * deactivate (e.g., log out) users after a period of inactivity

    The methods for managing these values deal with the timeouts as
    number-of-minutes. The default values and the methods are:

      * 20  modifiedUserTimeout()  setModifiedUserTimeout()
      * 20  cachedUserTimeout()    setCachedUserTimeout()
      * 20  activeUserTimeout()    setActiveUserTimeout()

    Subclasses of UserManager provide persistence such as to the file
    system or a MiddleKit store. Subclasses must implement all methods
    that raise AbstractErrors. Subclasses typically override (while still
    invoking super) addUser().

    Subclasses should ensure "uniqueness" of users. For example, invoking
    any of the userForSomething() methods repeatedly should always return
    the same user instance for a given key. Without uniqueness, consistency
    issues could arise with users that are modified.

    Please read the method docstrings and other class documentation to
    fully understand UserKit.
    """

    # region Init

    def __init__(self, userClass=None):
        if userClass is None:
            self._userClass = None
        else:
            self.setUserClass(userClass)
        self._cachedUsers = []
        self._cachedUsersBySerialNum = {}
        self.setModifiedUserTimeout(20)
        self.setCachedUserTimeout(20)
        self.setActiveUserTimeout(20)
        self._numActive = 0

    def shutDown(self):
        """Perform any tasks necessary to shut down the user manager.

        Subclasses may override and must invoke super as their *last* step.
        """
        # base method does nothing

    # endregion Init

    # region Settings

    def userClass(self):
        """Return the userClass, which is used by createUser.

        The default value is UserKit.User.User.
        """
        if self._userClass is None:
            self.setUserClass(User)
        return self._userClass

    def setUserClass(self, userClass):
        """Set the userClass, which cannot be None and must inherit from User.

        See also: userClass().
        """
        if not issubclass(userClass, User):
            raise TypeError(f'{userClass} is not a User class')
        self._userClass = userClass

    def modifiedUserTimeout(self):
        return self._modifiedUserTimeout

    def setModifiedUserTimeout(self, value):
        self._modifiedUserTimeout = value

    def cachedUserTimeout(self):
        return self._cachedUserTimeout

    def setCachedUserTimeout(self, value):
        self._cachedUserTimeout = value

    def activeUserTimeout(self):
        return self._activeUserTimeout

    def setActiveUserTimeout(self, value):
        self._activeUserTimeout = value

    # endregion Settings

    # region Basic user access

    def createUser(self, name, password, userClass=None):
        """Return a newly created user that is added to the manager.

        If userClass is not specified, the manager's default user class
        is instantiated. This not imply that the user is logged in.
        This method invokes self.addUser().

        See also: userClass(), setUserClass()
        """
        if userClass is None:
            userClass = self.userClass()
        user = userClass(manager=self, name=name, password=password)
        self.addUser(user)
        return user

    def addUser(self, user):
        if not isinstance(user, User):
            raise TypeError(f'{user} is not a User object')
        self._cachedUsers.append(user)
        if user.serialNum() in self._cachedUsersBySerialNum:
            raise KeyError('a user with this number already exists')
        self._cachedUsersBySerialNum[user.serialNum()] = user

    def userForSerialNum(self, serialNum, default=NoDefault):
        """Return the user with the given serialNum.

        The user record is pulled into memory if needed.
        """
        raise AbstractError(self.__class__)

    def userForExternalId(self, externalId, default=NoDefault):
        """Return the user with the given external id.

        The user record is pulled into memory if needed.
        """
        raise AbstractError(self.__class__)

    def userForName(self, name, default=NoDefault):
        """Return the user with the given name.

        The user record is pulled into memory if needed.
        """
        raise AbstractError(self.__class__)

    def users(self):
        """Return a list of all users (regardless of login status)."""
        raise AbstractError(self.__class__)

    def numActiveUsers(self):
        """Return the number of active users (e.g. the logged in users)."""
        return self._numActive

    def activeUsers(self):
        """Return a list of all active users."""
        raise AbstractError(self.__class__)

    def inactiveUsers(self):
        raise AbstractError(self.__class__)

    # endregion Basic user access

    # region Logging in and out

    def login(self, user, password):
        """Return the user if login is successful, otherwise return None."""
        if not isinstance(user, User):
            raise TypeError(f'{user} is not a User object')
        result = user.login(password, fromMgr=True)
        if result:
            self._numActive += 1
        return result

    def logout(self, user):
        if not isinstance(user, User):
            raise TypeError(f'{user} is not a User object')
        user.logout(fromMgr=True)
        self._numActive -= 1

    def loginSerialNum(self, serialNum, password):
        # pylint: disable=assignment-from-no-return
        user = self.userForSerialNum(serialNum, None)
        if user:
            return self.login(user, password)
        return None

    def loginExternalId(self, externalId, password):
        # pylint: disable=assignment-from-no-return
        user = self.userForExternalId(externalId, None)
        if user:
            return self.login(user, password)
        return None

    def loginName(self, userName, password):
        # pylint: disable=assignment-from-no-return
        user = self.userForName(userName, None)
        if user:
            return self.login(user, password)
        return None

    # endregion Logging in and out

    # region Cached

    def clearCache(self):
        """Clear the cache of the manager.

        Use with extreme caution. If your program maintains a reference
        to a user object, but the manager loads in a new copy later on,
        then consistency problems could occur.

        The most popular use of this method is in the regression test suite.
        """
        self._cachedUsers = []
        self._cachedUsersBySerialNum = {}

    # endregion Cached



"""JSON-RPC demo client contributed by Christoph Zwerschke"""

from .ExamplePage import ExamplePage


class JSONRPCClient(ExamplePage):
    """Demo client for using the JSON-RPC example."""

    def writeJavaScript(self):
        ExamplePage.writeJavaScript(self)
        self.write('''\
            <script src="jsonrpc.js"></script>
            <script>
            jsonrpc = new JSONRpcClient("JSONRPCExample");
            methods = jsonrpc.system.listMethods();
            function dojsonrpc() {
                p = document.getElementById("parameter").value;
                i = document.getElementById("method").selectedIndex;
                r = jsonrpc[methods[i]](p);
                document.getElementById("result").value = r;
            }
            </script>
        ''')

    def writeContent(self):
        self.write('''\
            <h3>JSON-RPC Example</h3>
            <p>This example shows how you can call methods
            of a <a href="https://www.jsonrpc.org/">JSON-RPC</a> servlet
            built with Webware for Python from your web browser
            via JavaScript (which has to be activated to run this demo).
            <noscript><span style="color:red">
            Unfortunately, JavaScript is not activated.
            </span></noscript></p>
            <p>The example uses a JSON-RPC JavaScript client
            based on Jan-Klaas' "JavaScript o lait" library.</p>
            <p>Type in any example text to be used as input parameter,
            choose one of the available methods to be invoked by the
            example servlet and press the button to display the result.</p>
            <table>
            <tr style="text-align:center">
            <th>Input parameter</th>
            <th>Remote method</th>
            <th>Result</th>
            </tr><tr style="text-align:center">
            <td>
            <input id="parameter" type="text" size="20" value="Hello, World!">
            </td><td>
            <select id="method" size="1">
            <script>
            for (m in methods)
                document.writeln('<option>' + methods[m] + '</option>');
            </script>
            </select>
            </td><td>
            <input type="text" size="20" id="result">
            </td>
            </tr><tr style="text-align:center">
            <td colspan="3">
            <input type="button" value="Invoke remote method"
                onclick="dojsonrpc()">
            </td>
            </tr>
            </table>
        ''')



import unittest

from MiscUtils.DateParser import parseDate, parseTime, parseDateTime


class TestDateTimeParser(unittest.TestCase):

    def testReturnType(self):
        from datetime import datetime
        self.assertIsInstance(
            parseDateTime('Mon Jul 21 02:56:20 1969'), datetime)

    def assertParses(self, s):
        self.assertEqual(parseDateTime(s).isoformat(), '1969-07-21T02:56:20')

    def testDefaultFormat(self):
        self.assertParses('Mon Jul 21 02:56:20 1969')

    def testCookieFormat(self):
        self.assertParses('Mon, 21-Jul-1969 02:56:20')

    def testISOFormat(self):
        self.assertParses('1969-07-21T02:56:20')

    def testShortISOFormat(self):
        self.assertParses('19690721T02:56:20')

    def testWrongFormat(self):
        self.assertRaises(ValueError,
                          parseDateTime, 'Mon Jul 21 02:56:20 19691')
        self.assertRaises(ValueError,
                          parseDateTime, '19691')

    def testDateUtilParser(self):
        try:
            # pylint: disable=unused-import
            from dateutil.parser import parse  # noqa: F401
        except ImportError:
            hasDateUtil = False
        else:
            hasDateUtil = True
        testString = 'July 21, 1969, 2:56:20'
        if hasDateUtil:
            self.assertParses(testString)
        else:
            self.assertRaises(ValueError, parseDateTime, testString)


class TestDateParser(unittest.TestCase):

    def testReturnType(self):
        from datetime import date
        self.assertIsInstance(parseDate('Mon Jul 21 02:56:20 1969'), date)

    def assertParses(self, s):
        self.assertEqual(parseDate(s).isoformat(), '1969-07-21')

    def testDefaultFormat(self):
        self.assertParses('Mon Jul 21 02:56:20 1969')

    def testCookieFormat(self):
        self.assertParses('Mon, 21-Jul-1969 02:56:20')

    def testISOFormat(self):
        self.assertParses('1969-07-21T02:56:20')

    def testShortISOFormat(self):
        self.assertParses('19690721T02:56:20')
        self.assertParses('1969-07-21')
        self.assertParses('19690721')

    def testWrongFormat(self):
        self.assertRaises(ValueError,
                          parseDateTime, 'Mon Jul 21 02:56:20 19691')
        self.assertRaises(ValueError,
                          parseDateTime, '19691')


class TestTimeParser(unittest.TestCase):

    def testReturnType(self):
        from datetime import time
        self.assertIsInstance(parseTime('Mon Jul 21 02:56:20 1969'), time)

    def assertParses(self, s):
        self.assertEqual(parseTime(s).isoformat(), '02:56:20')

    def testDefaultFormat(self):
        self.assertParses('Mon Jul 21 02:56:20 1969')

    def testCookieFormat(self):
        self.assertParses('Mon, 21-Jul-1969 02:56:20')

    def testISOFormat(self):
        self.assertParses('1969-07-21T02:56:20')

    def testShortISOFormat(self):
        self.assertParses('02:56:20')

    def testWrongFormat(self):
        self.assertRaises(ValueError,
                          parseDateTime, 'Mon Jul 21 02:65:20 1969')
        self.assertRaises(ValueError,
                          parseDateTime, '02:65')



import os
from time import time, localtime, gmtime, asctime

from .AdminSecurity import AdminSecurity


class Main(AdminSecurity):

    def title(self):
        return 'Admin'

    def writeContent(self):
        self.curTime = time()
        self.writeGeneralInfo()
        self.writeSignature()

    def writeGeneralInfo(self):
        app = self.application()
        info = (
            ('Webware Version', app.webwareVersionString()),
            ('Local Time', asctime(localtime(self.curTime))),
            ('Up Since', asctime(localtime(app.startTime()))),
            ('Num Requests', app.numRequests()),
            ('Working Dir', os.getcwd()),
            ('Active Sessions', len(app.sessions()))
        )
        self.writeln('''
<h2 style="text-align:center">Webware Administration Pages</h2>
<table style="margin-left:auto;margin-right:auto" class="NiceTable">
<tr class="TopHeading"><th colspan="2">Application Info</th></tr>''')
        for label, value in info:
            self.writeln(
                f'<tr><th style="text-align:left">{label}:</th>'
                f'<td>{value}</td></tr>')
        self.writeln('</table>')

    def writeSignature(self):
        self.writeln(f'''
<!--
begin-parse
{{
'Version': {self.application().webwareVersion()!r},
'LocalTime': {localtime(self.curTime)!r},
'GlobalTime': {gmtime(self.curTime)!r}
}}
end-parse
-->''')



"""Test Webware server script"""

import unittest

from os import environ, getcwd, chdir
from re import compile as reCompile
from subprocess import Popen, PIPE, STDOUT
from queue import Empty
from multiprocessing import Process, Queue, SimpleQueue
from urllib.request import urlopen

import psutil  # pylint: disable=import-error

expectedServerOutput = r'''
Webware for Python .+ Application

Process id: \d+
Start date/time: .+
Python: .+

ActivityLogColumns              = \[.+\]
ActivityLogFilename             = '.+'
AdminPassword                   = ''
AlwaysSaveSessions              = True
CacheServletClasses             = False
CacheServletInstances           = False
LogActivity                     = False
LogErrors                       = True
Verbose                         = True
WSGIWrite                       = True
WebwarePath                     = '.+'

Loading context: Examples at .+Examples
Loading context: Admin at .+Admin
Loading context: Testing at .+Testing
Plug-ins list: MiscUtils, WebUtils, TaskKit, UserKit, PSP
Loading plug-in: MiscUtils at .+MiscUtils
Loading plug-in: WebUtils at .+WebUtils
Loading plug-in: TaskKit at .+TaskKit
Loading plug-in: UserKit at .+UserKit
Loading plug-in: PSP at .+PSP
Loading context: PSP/Examples at .+Examples

Waitress serving Webware application...
Serving on http://.+:8080
'''

expectedStartPage = r'''
<!DOCTYPE html>
<html lang="en">
<head>
\s*<title>Welcome</title>
\s*<meta charset="utf-8">
<style>
html {
}
h1 { .+; }
#Page {
\s*display: table;
}
</style>
</head>
<body style="color:black;background-color:white">
<div id="Page"><div>
<div id="CornerTitle">Webware Examples</div><div id="Banner">Welcome</div>
</div><div id="Content">
<h2>Welcome to Webware .+!</h2>
</div></div></div>
</body>
</html>
'''


class RunServer(Process):
    """Run a given command until a given line is found in the output."""

    def __init__(self, cmd=None, waitForLine='Serving on '):
        super().__init__()
        self.cmd = cmd or ['webware', 'serve']
        self.waitForLine = waitForLine
        self.outputQueue = SimpleQueue()
        self.pollQueue = Queue()
        self.stopQueue = SimpleQueue()

    def getOutput(self):
        output = []
        while not self.outputQueue.empty():
            output.append(self.outputQueue.get())
        return output

    def getExitCode(self, timeout=10):
        try:
            return self.pollQueue.get(timeout=timeout)
        except Empty:
            self.stop()
            return f'timeout - {self.waitForLine!r} not found'

    def stop(self, timeout=10):
        if not self.is_alive():
            return
        # Because on Windows serve.exe starts another Python process,
        # we need to make sure the grandchildren are stopped as well.
        children = psutil.Process(self.pid).children(recursive=True)
        for p in children:
            try:
                p.terminate()
            except psutil.NoSuchProcess:
                pass
        children = psutil.wait_procs(children, timeout=timeout)[1]
        for p in children:
            try:
                p.kill()
            except psutil.NoSuchProcess:
                pass

    def run(self):
        # Because on Windows serve.exe starts another Python process,
        # we need to make sure that process runs in unbuffered mode as well.
        environ['PYTHONUNBUFFERED'] = '1'
        with Popen(self.cmd, bufsize=1, universal_newlines=True,
                   encoding='utf-8', stdout=PIPE, stderr=STDOUT) as p:
            outputStarted = False
            while True:
                ret = p.poll()
                if ret is not None:
                    break
                line = p.stdout.readline().rstrip()
                if line:
                    outputStarted = True
                if outputStarted:
                    self.outputQueue.put(line)
                    if line.startswith(self.waitForLine):
                        break
            self.pollQueue.put(ret)
            if ret is None:
                while True:
                    ret = p.poll()
                    if ret is not None:
                        break
                    line = p.stdout.readline().rstrip()
                    self.outputQueue.put(line)
            self.pollQueue.put(ret)


class TestServer(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.currentDir = getcwd()
        import webware
        webwareDir = webware.__path__[0]
        chdir(webwareDir)
        from Properties import version
        from MiscUtils.PropertiesObject import versionString
        cls.version = versionString(version)
        cls.server = RunServer()
        cls.server.start()
        cls.exitCode = cls.server.getExitCode()
        cls.running = cls.exitCode is None
        cls.output = cls.server.getOutput()
        cls.running = cls.exitCode is None

    @classmethod
    def tearDownClass(cls):
        try:
            cls.server.stop()
        finally:
            chdir(cls.currentDir)

    def getPage(self, path=''):
        return urlopen(f'http://localhost:8080/{path}')

    def compareOutput(self, gotLines, expectedLines):
        self.assertTrue(gotLines, 'We could not collect any output.')
        index = 0
        for expectedLine in expectedLines:
            expectedPattern = reCompile(expectedLine)
            foundLine = False
            lastIndex = index
            while True:
                try:
                    line = gotLines[index]
                except IndexError:
                    break
                index += 1
                if expectedPattern.match(line):
                    foundLine = True
                    break
            if not foundLine:
                break
        else:
            expectedLine = False
        nextOutput = '\n'.join(gotLines[lastIndex:][:3])
        fullOutput = '\n'.join(gotLines)
        msg = (f'Expected line not found:\n{expectedLine}\n\n'
               f'Next three lines were:\n{nextOutput}\n\n'
               f'Complete output was:\n{fullOutput}\n')
        self.assertTrue(foundLine, msg)

    def testExitCode(self):
        ret = self.exitCode
        self.assertIsNone(ret, f'Server exited with {ret}')

    def testExpectedServerOutput(self):
        expectedOutput = expectedServerOutput.strip().splitlines()
        self.compareOutput(self.output, expectedOutput)
        startLine = f'Webware for Python {self.version} Application'
        self.assertEqual(self.output[0], startLine)
        self.assertTrue(not any(
            'ERROR' in line or 'WARNING' in line for line in self.output))

    def testStartPage(self):
        self.assertTrue(self.running)
        expectedPage = expectedStartPage.strip().splitlines()
        with self.getPage() as page:
            page = page.read().decode('utf-8')
            page = [page.lstrip() for page in page.splitlines()]
            self.compareOutput(page, expectedPage)
            welcomeLine = f'Welcome to Webware for Python {self.version}'
            self.assertTrue(any(welcomeLine in line for line in page))

    def testShowTimeExample(self):
        self.assertTrue(self.running)
        with self.getPage('ShowTime') as page:
            page = page.read().decode('utf-8')
            self.assertIn('The current time is:', page)

    def testForward(self):
        self.assertTrue(self.running)
        with self.getPage('Forward') as page:
            page = page.read().decode('utf-8')
            self.assertIn('the <em>Forward</em> servlet speaking', page)
            self.assertIn('<h1>Hello from IncludeMe</h1>', page)

    def testIfModifiedSince(self):
        self.assertTrue(self.running)
        with self.getPage('Testing/TestIMS') as page:
            page = page.read().decode('utf-8')
            self.assertIn('Test If-Modified-Since support in Webware', page)
            self.assertIn('TestIMS passed.', page)



"""Test Webware Admin context"""

import unittest

from .AppTest import AppTest


class TestAdminLogin(AppTest, unittest.TestCase):

    settings = dict(
        PrintConfigAtStartUp=False,
        AdminPassword='LoginTestPassword'
    )

    def testLogin(self):
        r = self.testApp.get('/')
        r = r.click('Admin', index=0)
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Admin/')
        r.mustcontain(
            '<title>LoginPage</title>',
            'Please log in to view Administration Pages.',
            no=['>Webware Administration Pages<',
                'Login failed', 'Try again',
                'You have been logged out'])
        r = r.form.submit('login')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain('Login failed. Please try again.')
        r.form['username'] = 'admin'
        r.form['password'] = 'BadPassword'
        r = r.form.submit('login')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain('Login failed. Please try again.')
        r.form['username'] = 'alice'
        r.form['password'] = self.settings['AdminPassword']
        r = r.form.submit('login')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain('Login failed. Please try again.')
        r.form['username'] = 'admin'
        r.form['password'] = self.settings['AdminPassword']
        r = r.form.submit('login')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Admin</title>',
            '>Webware Administration Pages<',
            no=['Please log in', 'Login failed', 'Try again'])
        r = r.goto('./')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Admin</title>',
            '>Webware Administration Pages<',
            no=['Please log in', 'Login failed', 'Try again'])
        r = r.click('Logout')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>LoginPage</title>',
            'You have been logged out.',
            'Please log in to view Administration Pages.',
            no=['>Webware Administration Pages<',
                'Login failed', 'Try again'])
        r = r.goto('./')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>LoginPage</title>',
            'Please log in to view Administration Pages.',
            no=['>Webware Administration Pages<',
                'You have been logged out.',
                'Login failed', 'Try again'])


class TestAdminWithoutPassword(AppTest, unittest.TestCase):

    def testLogin(self):
        r = self.testApp.get('/Admin/')
        r.mustcontain(
            'Logins to admin pages are disabled'
            ' until you supply an AdminPassword in Application.config.')


class TestAdmin(AppTest, unittest.TestCase):

    settings = dict(
        PrintConfigAtStartUp=False,
        AdminPassword='AdminTestPassword',
        CacheServletClasses=True,
        CacheServletInstances=True
    )

    def setUp(self):
        AppTest.setUp(self)
        self.testApp.reset()
        r = self.testApp.get('/Admin/')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain('Please log in')
        r.form['username'] = 'admin'
        r.form['password'] = self.settings['AdminPassword']
        r = r.form.submit('login')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(no=['Login failed'])

    def tearDown(self):
        try:
            r = self.testApp.get('/Admin/?logout=yes')
            self.assertEqual(r.status, '200 OK')
            r.mustcontain('You have been logged out.')
        finally:
            AppTest.tearDown(self)

    def testStartPage(self):
        r = self.testApp.get('/Admin/')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Admin</title>',
            '>Webware Administration Pages<',
            no=['Please log in', 'Login failed'])
        r.mustcontain(
            '<th colspan="2">Application Info</th>',
            'Webware Version', 'Local Time', 'Up Since',
            'Num Requests', 'Working Dir', 'Active Sessions')

    def testActivityLog(self):
        r = self.testApp.get('/Admin/').click('Activity log')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Admin/Access')
        r.mustcontain('<title>View Activity</title>')

    def testErrorLog(self):
        r = self.testApp.get('/Admin/').click('Error log')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Admin/Errors')
        r.mustcontain('<title>View Errors</title>')

    def testConfig(self):
        r = self.testApp.get('/Admin/').click('Config')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Admin/Config')
        adminPassword = self.settings['AdminPassword']
        r.mustcontain(
            '<title>Config</title>',
            '<tr class="TopHeading"><th colspan="2">Application</th></tr>',
            f'<tr><th>AdminPassword</th><td>{adminPassword}</td></tr>',
            '<tr><th>LogActivity</th><td>False</td></tr>',
            '<tr><th>Verbose</th><td>True</td></tr>',
            '<tr><th>WSGIWrite</th><td>True</td></tr>')

    def testPlugins(self):
        r = self.testApp.get('/Admin/').click('Plug-ins')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Admin/PlugIns')
        webwareVersion = self.app.webwareVersionString()
        r.mustcontain(
            '<title>PlugIns</title>',
            'The following Plug-ins were found:',
            '<tr class="TopHeading"><th colspan="3">Plug-ins</th></tr>',
            '<tr class="SubHeading"><th>Name</th>'
            '<th>Version</th><th>Directory</th></tr>',
            f'<td style="text-align:center">{webwareVersion}</td>',
            '<td>MiscUtils</td>', '<td>PSP</td>', '<td>TaskKit</td>',
            '<td>UserKit</td>', '<td>WebUtils</td>')

    def testServletCache(self):
        r = self.testApp.get('/Admin/').click('Servlet Cache')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Admin/ServletCache')
        r.mustcontain(
            '<title>Servlet Cache</title>', '<h4>PythonServletFactory</h4>',
            '<p>Uniqueness: file</p>', "<p>Extensions: '.py', '.pyc'</p>",
            'Click any link to jump to the details for that path.', 'Flush',
            '<h5>Filenames:</h5>', '<tr><th>File</th><th>Directory</th></tr>',
            '<h5>Full paths:</h5>', '<tr><th>Servlet path</th></tr>',
            '>Main.py</a></td><td>', '>ServletCache.py</a></td><td>',
            '<h5>Details:</h5>', 'free reusable servlets:',
            '<strong>Main.py</strong>', '<strong>ServletCache.py</strong>',
            '<th>class</th>', '<th>instances</th>',
            '<th>mtime</th>', '<th>path</th>',
            '<td>Main</td>', '<td>ServletCache</td>',
            no='has been flushed')
        r = r.form.submit('flush_PythonServletFactory')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Servlet Cache</title>', '<h4>PythonServletFactory</h4>',
            'The servlet cache has been flushed.', 'Reload',
            no=['Click any link', 'Uniqueness:', '<h5>Filenames:</h5>'])
        r = r.form.submit()
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Servlet Cache</title>', '<h4>PythonServletFactory</h4>',
            '<p>Uniqueness: file</p>', "<p>Extensions: '.py', '.pyc'</p>",
            'Unique paths in the servlet cache: <strong>1</strong>',
            'Click any link to jump to the details for that path.',
            '<strong>ServletCache.py</strong>', '<td>ServletCache</td>',
            no=['has been flushed', 'Main.py', '<td>Main</td>'])
        r = self.testApp.get('/Admin/').click('Servlet Cache')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Servlet Cache</title>', '<h4>PythonServletFactory</h4>',
            '<p>Uniqueness: file</p>', "<p>Extensions: '.py', '.pyc'</p>",
            'Unique paths in the servlet cache: <strong>2</strong>',
            'Click any link to jump to the details for that path.',
            '<strong>Main.py</strong>', '<td>Main</td>',
            '<strong>ServletCache.py</strong>', '<td>ServletCache</td>',
            no='has been flushed')

    def testAppControl(self):
        r = self.testApp.get('/Admin/').click('Application Control')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Admin/AppControl')
        r.mustcontain(
            '<title>AppControl</title>',
            'Clear the class and instance caches of each servlet factory.',
            'Reload the selected Python modules. Be careful!',
            '<input type="checkbox" name="reloads"'
            ' value="Admin.AdminPage"> Admin.AdminPage<br>')
        r = r.form.submit('action', index=0)
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>AppControl</title>',
            'Flushing cache of PythonServletFactory...',
            'The caches of all factories have been flushed.',
            'Click here to view the Servlet cache:',
            no=['<input type="checkbox"', 'Clear the class and instance'])
        r = r.click('Servlet Cache', index=1)
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Servlet Cache</title>', '<h4>PythonServletFactory</h4>',
            'Unique paths in the servlet cache: <strong>1</strong>')
        r = r.click('Application Control')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>AppControl</title>',
            'Clear the class and instance caches of each servlet factory.',
            'Reload the selected Python modules. Be careful!',
            '<input type="checkbox" name="reloads"'
            ' value="Admin.AdminPage"> Admin.AdminPage<br>')
        for i, checkbox in enumerate(
                r.html.find_all('input', attrs={'name': 'reloads'})):
            if checkbox.attrs['value'] == 'Admin.AdminPage':
                r.form.get('reloads', index=i).checked = True
                break
        else:
            self.fail('Did not find expected checkbox for Admin.AdminPage')
        r = r.form.submit('action', index=1)
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>AppControl</title>',
            'Reloading selected modules.',
            "Reloading &lt;module 'Admin.AdminPage'",
            'The selected modules have been reloaded.',
            no=['<input type="checkbox"', 'Be careful!'])
        r = r.goto('/Admin/AppControl')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>AppControl</title>',
            'Clear the class and instance caches of each servlet factory.',
            'Reload the selected Python modules. Be careful!',
            '<input type="checkbox" name="reloads"'
            ' value="Admin.AdminPage"> Admin.AdminPage<br>')



import os

from Page import Page

try:
    from pygments import highlight
    from pygments.lexers import PythonLexer
    from pygments.formatters import HtmlFormatter
except ImportError:
    highlight = None

errorPage = """\
<!DOCTYPE html>
<html>
<head>
<meta charset="utf-8">
<title>Error</title>
<style>
body {{ font-family:sans-serif; font-size:16px; }}
</style>
</head>
<body>
<h3 style="color:#a00">Error</h3>
<p>{}</p>
</body>
</html>
"""

viewPage = """\
<!DOCTYPE html>
<html>
<head>
<meta charset="utf-8">
<title>{}</title>
<style>{}</style>
</head>
<body class="highlight">
{}
</body>
</html>
"""


class Colorize(Page):
    """Syntax highlights Python source files."""

    def respond(self, transaction):
        """Write out a syntax highlighted version of the file."""
        res = transaction._response
        req = self.request()
        if not req.hasField('filename'):
            res.write(errorPage.format('No filename given to syntax color!'))
            return
        filename = req.field('filename')
        filename = self.request().serverSidePath(os.path.basename(filename))
        try:
            with open(filename) as f:
                source = f.read()
        except IOError:
            source = None
        if source is None:
            res.write(errorPage.format(
                f'The requested file {os.path.basename(filename)!r}'
                ' does not exist in the proper directory.'))
            return
        if highlight:
            formatter = HtmlFormatter()
            lexer = PythonLexer()
            output = highlight(source, lexer, formatter)
            style = formatter.get_style_defs('.highlight')
            output = viewPage.format(filename, style, output)
        else:
            print("Pygments highlighter not found - showing plain source")
            res.setHeader('Content-Type', 'text/plain')
            output = source
        res.write(output)



"""CSVParser.py

A parser for CSV files.
"""

# The states of the parser
(StartRecord, StartField, InField, QuoteInField,
 InQuotedField, QuoteInQuotedField, EndQuotedField) = range(7)

# State handlers can return Finished to terminate parsing early
Finished = 10


class ParseError(Exception):
    """CSV file parse error."""


class CSVParser:
    """Parser for CSV files.

    Parses CSV files including all subtleties such as:

    * commas in fields
    * double quotes in fields
    * embedded newlines in fields

    Examples of programs that produce such beasts include MySQL and Excel.

    For a higher-level, friendlier CSV class with many conveniences,
    see `DataTable` (which uses this class for its parsing).

    Example::

        records = []
        parse = CSVParser().parse
        for line in lines:
            results = parse(line)
            if results is not None:
                records.append(results)


    CREDIT

    The algorithm was taken directly from the open source Python
    C-extension, csv: https://www.object-craft.com.au/projects/csv/

    It would be nice to use the csv module when present, since it is
    substantially faster. Before that can be done, it needs to support
    `allowComments` and `stripWhitespace`, and pass the TestCSVParser.py
    test suite.
    """

    def __init__(
            self, allowComments=True, stripWhitespace=True, fieldSep=',',
            autoReset=True, doubleQuote=True):
        """Create a new CSV parser.

        `allowComments`:
          If true (the default), then comment lines using the Python comment
          marker are allowed.
        `stripWhitespace`:
          If true (the default), then left and right whitespace is stripped
          off from all fields.
        `fieldSep`:
          Defines the field separator string (a comma by default).
        `autoReset`:
          If true (the default), recover from errors automatically.
        `doubleQuote`:
          If true (the default), assume quotes in fields are escaped by
          appearing doubled.
        """
        # settings
        self._allowComments = allowComments
        self._stripWhitespace = stripWhitespace
        self._doubleQuote = doubleQuote
        self._fieldSep = fieldSep
        self._autoReset = autoReset

        # Other
        self._state = StartRecord
        self._fields = []
        self._hadParseError = False
        self._field = []  # a list of chars for the cur field
        self.addChar = self._field.append

        # The handlers for the various states
        self._handlers = [
            self.startRecord,
            self.startField,
            self.inField,
            self.quoteInField,
            self.inQuotedField,
            self.quoteInQuotedField,
            self.endQuotedField,
        ]

    # region Parse

    def parse(self, line):
        """Parse a single line and return a list of string fields.

        Returns None if the CSV record contains embedded newlines and
        the record is not yet complete.
        """
        if self._autoReset and self._hadParseError:
            self.reset()
        handlers = self._handlers

        i = 0
        lineLen = len(line)
        while i < lineLen:
            c = line[i]
            if c == '\r':
                i += 1
                if i == lineLen:
                    break  # Mac end of line
                c = line[i]
                if c == '\n':
                    i += 1
                    if i == lineLen:
                        break  # Win end of line

                self._hadParseError = True
                raise ParseError('Newline inside string')

            if c == '\n':
                i += 1
                if i == lineLen:
                    break  # Unix end of line

                self._hadParseError = True
                raise ParseError('Newline inside string')

            if handlers[self._state](c) == Finished:
                break  # process a character

            i += 1

        handlers[self._state]('\0')  # signal the end of the input

        if self._state == StartRecord:
            fields = self._fields
            self._fields = []
            if self._stripWhitespace:
                fields = [field.strip() for field in fields]
            return fields

        return None  # indicates multi-line record; e.g. not finished

    # endregion Parse

    # region Reset

    def reset(self):
        """Reset the parser.

        Resets the parser to a fresh state in order to recover from
        exceptions. But if autoReset is true (the default), this is
        done automatically.
        """
        self._fields = []
        self._state = StartRecord
        self._hadParseError = False

    # endregion Reset

    # region State Handlers

    def startRecord(self, c):
        if c != '\0':  # not empty line
            if c == '#' and self._allowComments:
                return Finished
            self._state = StartField
            self.startField(c)

    def startField(self, c):
        if c == '"':
            self._state = InQuotedField  # start quoted field
        elif c == self._fieldSep:
            self.saveField()  # save empty field
        elif c == ' ' and self._stripWhitespace:
            pass  # skip over preceding whitespace
        elif c == '\0':
            self.saveField()  # save empty field
            self._state = StartRecord
        else:
            self.addChar(c)  # begin new unquoted field
            self._state = InField

    def inField(self, c):
        # in unquoted field
        if c == self._fieldSep:
            self.saveField()
            self._state = StartField
        elif c == '\0':
            self.saveField()  # end of line
            self._state = StartRecord
        elif c == '"' and self._doubleQuote:
            self._state = QuoteInField
        else:
            self.addChar(c)  # normal character

    def quoteInField(self, c):
        self.addChar('"')
        if c == '"':
            self._state = InField  # save "" as "
        elif c == '\0':
            self.saveField()  # end of line
            self._state = StartRecord
        elif c == self._fieldSep:
            self.saveField()
            self._state = StartField
        else:
            self.addChar(c)  # normal character
            self._state = InField

    def inQuotedField(self, c):
        if c == '"':
            if self._doubleQuote:
                self._state = QuoteInQuotedField
            else:
                self.saveField()  # end of field
                self._state = EndQuotedField
        elif c == '\0':
            self.addChar('\n')  # end of line
        else:
            self.addChar(c)  # normal character

    def quoteInQuotedField(self, c):
        if c == '"':
            self.addChar('"')  # save "" as "
            self._state = InQuotedField
        elif c == self._fieldSep:
            self.saveField()
            self._state = StartField
        elif c == ' ' and self._stripWhitespace:
            pass  # skip it
        elif c == '\0':
            self.saveField()  # end of line
            self._state = StartRecord
        else:
            self._hadParseError = True  # illegal
            raise ParseError(f'{self._fieldSep} expected after "')

    def endQuotedField(self, c):
        if c == self._fieldSep:  # seen closing " on quoted field
            self._state = StartField  # wait for new field
        elif c == '\0':
            self._state = StartRecord  # end of line
        else:
            self._hadParseError = True
            raise ParseError(f'{self._fieldSep} expected after "')

    def saveField(self):
        self._fields.append(''.join(self._field))
        self._field = []
        self.addChar = self._field.append

    # endregion State Handlers


# Call the global function parse() if you like the default settings
_parser = CSVParser()
parse = _parser.parse



from .ExamplePage import ExamplePage


class CustomError(Exception):
    pass


class Error(ExamplePage):

    def title(self):
        return 'Error raising Example'

    def writeContent(self):
        error = self.request().field('error', None)
        if error:
            msg = 'You clicked that button!'
            if error.startswith('String'):
                error = msg
            elif error.startswith('Custom'):
                error = CustomError(msg)
            elif error.startswith('System'):
                error = SystemError(msg)
            else:
                error = RuntimeError(msg)
            self.writeln('<p>About to raise an error...</p>')
            raise error
        self.writeln('''<h1>Error Test</h1>
<form action="Error" method="post">
<p><select name="error" size="1">
<option selected>Runtime Error</option>
<option>System Error</option>
<option>Custom Error</option>
</select>
<input type="submit" value="Don't click this button!"></p>
</form>''')



"""Mock Webware Session class."""

from time import time


class Session:
    """Mock session."""

    _lastExpired = None

    def __init__(self, identifier=7, value=None):
        self._identifier = f'foo-{identifier}'
        self._data = dict(bar=value or identifier * 6)
        self._expired = self._dirty = False
        self._timeout = 1800
        self._lastAccessTime = time() - identifier * 400
        self._isNew = True

    def identifier(self):
        return self._identifier

    def expiring(self):
        self._expired = True
        Session._lastExpired = self._identifier

    def isDirty(self):
        return self._dirty

    def setDirty(self, dirty=True):
        self._dirty = dirty

    def isExpired(self):
        return self._expired

    def timeout(self):
        return self._timeout

    def data(self):
        return self._data

    def bar(self):
        return self._data.get('bar')

    def setBar(self, value):
        self._isNew = False
        self._data['bar'] = value

    def isNew(self):
        return self._isNew

    def lastAccessTime(self):
        return self._lastAccessTime



"""DictForArgs.py

See the doc string for the dictForArgs() function.

Also, there is a test suite in Tests/TestDictForArgs.py
"""

import re


class DictForArgsError(Exception):
    """Error when building dictionary from arguments."""


def _SyntaxError(s):
    raise DictForArgsError('Syntax error: %r' % s)


_nameRE = re.compile(r'\w+')
_equalsRE = re.compile(r'\=')
_stringRE = re.compile(r'''"[^"]+"|'[^']+'|\S+''')
_whiteRE = re.compile(r'\s+')

_REs = [_nameRE, _equalsRE, _stringRE, _whiteRE]


def dictForArgs(s):
    """Build dictionary from arguments.

    Takes an input such as::

        x=3
        name="foo"
        first='john' last='doe'
        required border=3

    And returns a dictionary representing the same. For keys that aren't
    given an explicit value (such as 'required' above), the value is '1'.

    All values are interpreted as strings. If you want ints and floats,
    you'll have to convert them yourself.

    This syntax is equivalent to what you find in HTML and close to other
    ML languages such as XML.

    Returns {} for an empty string.

    The informal grammar is::

        (NAME [=NAME|STRING])*

    Will raise DictForArgsError if the string is invalid.

    See also: pyDictForArgs() and expandDictWithExtras() in this module.
    """

    s = s.strip()

    # Tokenize

    verbose = False
    matches = []
    start = 0
    sLen = len(s)

    if verbose:
        print(f'>> dictForArgs({s!r})')
        print('>> sLen:', sLen)
    while start < sLen:
        for regEx in _REs:
            if verbose:
                print('>> try:', regEx)
            match = regEx.match(s, start)
            if verbose:
                print('>> match:', match)
            if match is not None:
                if match.re is not _whiteRE:
                    matches.append(match)
                start = match.end()
                if verbose:
                    print('>> new start:', start)
                break
        else:
            _SyntaxError(s)

    if verbose:
        names = []
        for match in matches:
            if match.re is _nameRE:
                name = 'name'
            elif match.re is _equalsRE:
                name = 'equals'
            elif match.re is _stringRE:
                name = 'string'
            elif match.re is _whiteRE:
                name = 'white'
            names.append(name)
        print('>> names =', names)

    # Process tokens

    # At this point we have a list of all the tokens (as re.Match objects)
    # We need to process these into a dictionary.

    d = {}
    matchesLen = len(matches)
    i = 0
    while i < matchesLen:
        match = matches[i]
        if i + 1 < matchesLen:
            peekMatch = matches[i+1]
        else:
            peekMatch = None
        if match.re is _nameRE:
            if peekMatch is not None:
                if peekMatch.re is _nameRE:
                    # We have a name without an explicit value
                    d[match.group()] = '1'
                    i += 1
                    continue
                if peekMatch.re is _equalsRE:
                    if i + 2 < matchesLen:
                        target = matches[i+2]
                        if target.re is _nameRE or target.re is _stringRE:
                            value = target.group()
                            if value[0] == "'" or value[0] == '"':
                                value = value[1:-1]
                            d[match.group()] = value
                            i += 3
                            continue
            else:
                d[match.group()] = '1'
                i += 1
                continue
        _SyntaxError(s)

    if verbose:
        print()

    return d


def pyDictForArgs(s):
    """Build dictionary from arguments.

    Takes an input such as::

        x=3
        name="foo"
        first='john'; last='doe'
        list=[1, 2, 3]; name='foo'

    And returns a dictionary representing the same.

    All values are interpreted as Python expressions. Any error in these
    expressions will raise the appropriate Python exception. This syntax
    allows much more power than dictForArgs() since you can include
    lists, dictionaries, actual ints and floats, etc.

    This could also open the door to hacking your software if the input
    comes from a tainted source such as an HTML form or an unprotected
    configuration file.

    Returns {} for an empty string.

    See also: dictForArgs() and expandDictWithExtras() in this module.
    """
    if s:
        s = s.strip()
    if not s:
        return {}

    # special case: just a name
    # meaning: name=1
    # example: isAbstract
    if ' ' not in s and '=' not in s and s[0].isalpha():
        s += '=1'

    results = {}
    exec(s, results)

    del results['__builtins__']
    return results


def expandDictWithExtras(d, key='Extras',
                         delKey=True, dictForArgs=dictForArgs):
    """Return a dictionary with the 'Extras' column expanded by dictForArgs().

    For example, given::

        {'Name': 'foo', 'Extras': 'x=1 y=2'}

    The return value is::

        {'Name': 'foo', 'x': '1', 'y': '2'}

    The key argument controls what key in the dictionary is used to hold
    the extra arguments. The delKey argument controls whether that key and
    its corresponding value are retained.
    The same dictionary may be returned if there is no extras key.
    The most typical use of this function is to pass a row from a DataTable
    that was initialized from a CSV file (e.g., a spreadsheet or tabular file).
    FormKit and MiddleKit both use CSV files and allow for an Extras column
    to specify attributes that occur infrequently.
    """
    if key in d:
        newDict = dict(d)
        if delKey:
            del newDict[key]
        newDict.update(dictForArgs(d[key]))
        return newDict
    return d


# old (deprecated) aliases
DictForArgs, PyDictForArgs, ExpandDictWithExtras = (
    dictForArgs, pyDictForArgs, expandDictWithExtras)



from MiscUtils.Configurable import Configurable, NoDefault


class ConfigurableForServerSidePath(Configurable):
    """Configuration file functionality incorporating a server side path.

    This is a version of `MiscUtils.Configurable.Configurable` that provides
    a customized `setting` method for classes which have a `serverSidePath`
    method. If a setting's name ends with ``Filename`` or ``Dir``, its value
    is passed through `serverSidePath` before being returned.

    In other words, relative filenames and directory names are expanded with
    the location of the object, *not* the current directory.

    Application is a prominent class that uses this mix-in. Any class that
    has a `serverSidePath` method and a `Configurable` base class, should
    inherit this class instead.

    This is used for `MakeAppWorkDir`, which changes the `serverSidePath`.
    """

    def setting(self, name, default=NoDefault):
        """Return setting, using the server side path when indicated.

        Returns the setting, filtered by `self.serverSidePath()`,
        if the name ends with ``Filename`` or ``Dir``.
        """
        value = Configurable.setting(self, name, default)
        if name.endswith(('Dir', 'Filename')) and (
                value or name.endswith('Dir')):
            value = self.serverSidePath(value)  # pylint: disable=no-member
        return value



import http.cookies as CookieEngine

from MiscUtils.Funcs import positiveId


class Cookie:
    """Delicious cookies.

    Cookie is used to create cookies that have additional attributes
    beyond their value.

    Note that web browsers don't typically send any information with
    the cookie other than its value. Therefore `HTTPRequest.cookie`
    simply returns a value such as an integer or a string.

    When the server sends cookies back to the browser, it can send a cookie
    that simply has a value, or the cookie can be accompanied by various
    attributes (domain, path, max-age, ...) as described in `RFC 2109`_.
    Therefore, in `HTTPResponse`, `setCookie` can take either an instance of
    the `Cookie` class, as defined in this module, or a value.

    Note that `Cookie` values get pickled (see the `pickle` module), so
    you can set and get cookies that are integers, lists, dictionaries, etc.

    .. _`RFC 2109`: ftp://ftp.isi.edu/in-notes/rfc2109.txt
    """

    # Future
    #
    # * This class should provide error checking in the setFoo()
    #   methods. Or maybe our internal Cookie implementation
    #   already does that?
    # * This implementation is probably not as efficient as it
    #   should be, [a] it works and [b] the interface is stable.
    #   We can optimize later.

    # region Init

    def __init__(self, name, value):
        """Create a cookie.

        Properties other than `name` and `value` are set with methods.
        """
        self._cookies = CookieEngine.SimpleCookie()
        self._name = name
        self._value = value
        self._cookies[name] = value
        self._cookie = self._cookies[name]

    def __repr__(self):
        return ('{}(id=0x{:x}, name={!r}, domain={!r},'
                ' path={!r}, value={!r}, expires={!r}, maxAge={!r}').format(
                    self.__class__.__name__, positiveId(self),
                    self.name(), self.domain(),
                    self.path(), self.value(), self.expires(), self.maxAge())

    # endregion Init

    # region Accessors

    def comment(self):
        return self._cookie['comment']

    def domain(self):
        return self._cookie['domain']

    def maxAge(self):
        return self._cookie['max-age']

    def expires(self):
        return self._cookie['expires']

    def name(self):
        return self._name

    def path(self):
        return self._cookie['path']

    def isSecure(self):
        return self._cookie['secure']

    def value(self):
        return self._value

    def version(self):
        return self._cookie['version']

    # endregion Accessors

    # region Setters

    def setComment(self, comment):
        self._cookie['comment'] = comment

    def setDomain(self, domain):
        self._cookie['domain'] = domain

    def setExpires(self, expires):
        self._cookie['expires'] = expires

    def setMaxAge(self, maxAge):
        self._cookie['max-age'] = maxAge

    def setPath(self, path):
        self._cookie['path'] = path

    def setSecure(self, secure=True):
        self._cookie['secure'] = secure

    def setValue(self, value):
        self._value = value
        self._cookies[self._name] = value

    def setVersion(self, version):
        self._cookie['version'] = version

    # endregion Setters

    # region Misc

    def delete(self):
        """Delete a cookie.

        When sent, this should delete the cookie from the user's
        browser, by making it empty, expiring it in the past,
        and setting its max-age to 0. One of these will delete
        the cookie for any browser (which one actually works
        depends on the browser).
        """
        self._value = ''
        self._cookie['expires'] = "Mon, 01-Jan-1900 00:00:00 GMT"
        self._cookie['max-age'] = 0

    def headerValue(self):
        """Return header value.

        Returns a string with the value that should be used
        in the HTTP headers.
        """
        values = list(self._cookies.values())
        if len(values) != 1:
            raise ValueError('Invalid cookie')
        return values[0].OutputString()

    # endregion Misc



"""The basic User class."""

from time import time

from MiscUtils.Funcs import uniqueId


class User:
    """The base class for a UserKit User."""

    # region Init

    def __init__(self, manager=None, name=None, password=None):
        self._creationTime = time()

        self._manager = None
        self._name = None
        self._password = None
        self._isActive = False
        self._externalId = None

        if name is not None:
            self.setName(name)
        if manager is not None:
            self.setManager(manager)
        if password is not None:
            self.setPassword(password)

    def __str__(self):
        return self.name()

    # endregion Init

    # region Attributes

    def manager(self):
        return self._manager

    def setManager(self, manager):
        """Set the manager, which can only be done once."""
        if self._manager is not None:
            raise KeyError('Manager has already been set')
        from .UserManager import UserManager
        if not isinstance(manager, UserManager):
            raise TypeError(f'{manager} is not a UserManager object')
        if manager.userForName(self.name(), None) is not None:
            raise KeyError(f'There is already a user named {self._name!r}.')
        self._manager = manager

    def serialNum(self):
        return self._serialNum

    def setSerialNum(self, serialNum):
        self._serialNum = serialNum

    def externalId(self):
        if self._externalId is None:
            self._externalId = uniqueId(self)
        return self._externalId

    def name(self):
        return self._name

    def setName(self, name):
        """Set the name, which can only be done once."""
        if self._name is not None:
            raise KeyError('name has already been set')
        self._name = name

    def password(self):
        return self._password

    def setPassword(self, password):
        self._password = password

    def isActive(self):
        return self._isActive

    def creationTime(self):
        return self._creationTime

    def lastAccessTime(self):
        return self._lastAccessTime

    def lastLoginTime(self):
        return self._lastLoginTime

    # endregion Attributes

    # region Log in and out

    def login(self, password, fromMgr=0):
        """Return self if the login is successful and None otherwise."""
        if not fromMgr:
            # Our manager needs to know about this.
            # So make sure we go through him.
            return self.manager().login(self, password)
        if password == self.password():
            self._isActive = True
            self._lastLoginTime = self._lastAccessTime = time()
            return self
        if self._isActive:
            # We were already logged in, but we tried again and
            # this time it failed. Logout:
            self.logout()
        return None

    def logout(self, fromMgr=False):
        if fromMgr:
            self._isActive = False
            self._lastLogoutTime = time()
        else:
            # Our manager needs to know about this.
            # So make sure we go through him.
            self.manager().logout(self)

    # region Log in and out

    # region Notifications

    def wasAccessed(self):
        self._lastAccessTime = time()

    # endregion Notifications



"""The RoleUserManager class."""

from .RoleUserManagerMixIn import RoleUserManagerMixIn
from .UserManager import UserManager


class RoleUserManager(UserManager, RoleUserManagerMixIn):
    """See the base classes for more information."""

    def __init__(self, userClass=None):
        UserManager.__init__(self, userClass)
        RoleUserManagerMixIn.__init__(self)



"""CSVJoiner.py

A helper function for joining CSV fields.
"""


def joinCSVFields(fields):
    """Create a CSV record by joining its fields.

    Returns a CSV record (e.g. a string) from a sequence of fields.
    Fields containing commands (,) or double quotes (") are quoted,
    and double quotes are escaped ("").
    The terminating newline is *not* included.
    """
    newFields = []
    for field in fields:
        if not isinstance(field, str):
            raise UnicodeDecodeError('CSV fields should be strings')
        if '"' in field:
            newField = '"{}"'.format(field.replace('"', '""'))
        elif ',' in field or '\n' in field or '\r' in field:
            newField = f'"{field}"'
        else:
            newField = field
        newFields.append(newField)
    return ','.join(newFields)



# pylint: disable=unused-import,wrong-import-order
from . import redis  # noqa: F401 (mock redis installation)

from SessionRedisStore import SessionRedisStore

from .TestSessionMemoryStore import TestSessionMemoryStore


class SessionRedisStoreTest(TestSessionMemoryStore):

    _storeClass = SessionRedisStore

    def setUp(self):
        TestSessionMemoryStore.setUp(self)

    def tearDown(self):
        TestSessionMemoryStore.tearDown(self)

    def testCleanStaleSessions(self):
        self._store.cleanStaleSessions()



import os
import unittest

from io import StringIO

from MiscUtils.DataTable import DataTable, DataTableError, TableColumn

cannotReadExcel = not DataTable.canReadExcel()


class TestTableColumn(unittest.TestCase):

    def testWithType(self):
        c = TableColumn('foo:int')
        self.assertEqual(c.name(), 'foo')
        self.assertTrue(c.type() is int)

    def testWithoutType(self):
        c = TableColumn('bar')
        self.assertEqual(c.name(), 'bar')
        self.assertTrue(c.type() is None)

    def testWrongSpec(self):
        self.assertRaises(DataTableError, TableColumn, 'foo:bar')
        self.assertRaises(DataTableError, TableColumn, 'foo:bar:baz')

    def testValueForRawValue(self):
        c = TableColumn('foo:int')
        self.assertEqual(c.valueForRawValue(''), 0)
        self.assertEqual(c.valueForRawValue('1'), 1)
        self.assertEqual(c.valueForRawValue(2), 2)
        self.assertEqual(c.valueForRawValue(2.5), 2)
        c = TableColumn('bar:str')
        self.assertEqual(c.valueForRawValue(''), '')
        self.assertEqual(c.valueForRawValue('1'), '1')
        self.assertEqual(c.valueForRawValue(2), '2')
        self.assertEqual(c.valueForRawValue('x'), 'x')
        c = TableColumn('bar:float')
        self.assertEqual(c.valueForRawValue(''), 0.0)
        self.assertEqual(c.valueForRawValue('1'), 1.0)
        self.assertEqual(c.valueForRawValue('1.5'), 1.5)
        self.assertEqual(c.valueForRawValue(2.5), 2.5)
        self.assertEqual(c.valueForRawValue(3), 3.0)


class Record:

    def __init__(self, **record):
        self.record = record

    def hasValueForKey(self, key):
        return key in self.record

    def valueForKey(self, key, default=None):
        return self.record.get(key, default)


class TestDataTable(unittest.TestCase):

    def _testSource(self, name, src, headings, data):
        dataTable = DataTable()
        lines = src.splitlines()
        dataTable.readLines(lines)
        self.assertEqual([c.name() for c in dataTable.headings()], headings)
        for i, values in enumerate(dataTable):
            match = data[i]
            asList = values.asList()
            self.assertEqual(
                asList, match,
                f'{name}: Row {i}: Expected {match!r}, but got {asList!r}')

    def testBasicWithPickle(self):
        DataTable.usePickleCache = True
        self._testBasic()
        self._testBasic()

    def testBasicWithoutPickle(self):
        DataTable.usePickleCache = False
        self._testBasic()

    def _testBasic(self):
        """Simple tests..."""

        # Create table
        t = DataTable()

        # Headings 1
        t = DataTable()
        t.setHeadings([
            TableColumn('name'), TableColumn('age:int'),
            TableColumn('rating:float')])

        # Headings 2
        t = DataTable()
        t.setHeadings(['name', 'age:int', 'rating:float'])

        # Adding and accessing data
        data = [
            ['John', '26', '7.25'],
            ['Mary', 32, 8.5],
            dict(name='Fred', age=28, rating=9.0),
            Record(name='Wilma', age=27, rating=9.5)
        ]
        for obj in data:
            t.append(obj)
        self.assertEqual(t[-4]['name'], 'John')
        self.assertEqual(t[-3]['name'], 'Mary')
        self.assertEqual(t[-2]['name'], 'Fred')
        self.assertEqual(t[-1]['name'], 'Wilma')
        self.assertEqual(
            t[-4].asDict(),
            {'name': 'John', 'age': 26, 'rating': 7.25})
        self.assertEqual(t[-3].asList(), data[-3])
        self.assertEqual(t[-2].asDict(), data[-2])
        self.assertEqual(t[-1].asList(), ['Wilma', 27, 9.5])

        # Printing
        # print(t)

        # Writing file (CSV)
        answer = '''\
name,age,rating
John,26,7.25
Mary,32,8.5
Fred,28,9.0
Wilma,27,9.5
'''
        out = StringIO()
        t.writeFile(out)
        results = out.getvalue()
        self.assertEqual(results, answer,
                         f'\n{results!r}\n{answer!r}\n')

        # Accessing rows
        for row in t:
            self.assertEqual(row['name'], row[0])
            self.assertEqual(row['age'], row[1])
            self.assertEqual(row['rating'], row[2])
            self.assertEqual(sum(1 for item in row), 3)

        # Default type
        t = DataTable(defaultType='int')
        t.setHeadings(list('xyz'))
        t.append([1, 2, 3])
        t.append([4, 5, 6])
        self.assertEqual(t[0]['x'] - t[1]['z'], -5)

    def testBasics(self):
        # Basics
        src = '''\
"x","y,y",z
a,b,c
a,b,"c,d"
"a,b",c,d
"a","b","c"
"a",b,"c"
"a,b,c"
"","",""
"a","",
'''
        headings = ['x', 'y,y', 'z']
        data = [
            ['a', 'b', 'c'],
            ['a', 'b', 'c,d'],
            ['a,b', 'c', 'd'],
            ['a', 'b', 'c'],
            ['a', 'b', 'c'],
            ['a,b,c', '', ''],
            ['', '', ''],
            ['a', '', '']
        ]
        self._testSource('Basics', src, headings, data)

        # Comments
        src = '''\
a:int,b:int
1,2
#3,4
5,6
'''
        headings = ['a', 'b']
        data = [
            [1, 2],
            [5, 6],
        ]
        self._testSource('Comments', src, headings, data)

        # Multiline records
        src = '''\
a
"""Hi
there"""
'''
        headings = ['a']
        data = [
            ['"Hi\nthere"'],
        ]
        self._testSource('Multiline records', src, headings, data)

        # MiddleKit enums
        src = '''\
Class,Attribute,Type,Extras
#Foo,
,what,enum,"Enums=""foo, bar"""
,what,enum,"Enums='foo, bar'"
'''
        headings = 'Class,Attribute,Type,Extras'.split(',')
        data = [
            ['', 'what', 'enum', 'Enums="foo, bar"'],
            ['', 'what', 'enum', "Enums='foo, bar'"],
        ]
        self._testSource('MK enums', src, headings, data)

        # Unfinished multiline record
        try:
            DataTable().readString('a\n"1\n')
        except DataTableError:
            pass  # just what we were expecting
        else:
            raise Exception(
                'Failed to raise exception for unfinished multiline record')

    def testDefaultUsePickleCache(self):
        t = DataTable()
        self.assertIs(t._usePickleCache, True)

    def testCsvWithPickle(self):
        self._testCsv(removePickleCache=False)
        self._testCsv()

    def testCsvWithoutPickle(self):
        self._testCsv(usePickleCache=False)

    def _testCsv(self, usePickleCache=True, removePickleCache=True):
        csvFile = os.path.join(os.path.dirname(__file__), 'Sample.csv')
        try:
            t = DataTable(csvFile, usePickleCache=usePickleCache)
        finally:
            pickleFile = csvFile + '.pickle.cache'
            self.assertIs(os.path.exists(pickleFile), usePickleCache)
            if usePickleCache and removePickleCache:
                os.remove(pickleFile)
        self.assertEqual(t[0][0], 'Video', t[0])

    @unittest.skipIf(cannotReadExcel, "Cannot read Excel files")
    def testExcelWithPickle(self):
        self._testExcel(removePickleCache=False)
        self._testExcel()

    @unittest.skipIf(cannotReadExcel, "Cannot read Excel files")
    def testExcelWithoutPickle(self):
        self._testExcel(usePickleCache=False)

    def _testExcel(self, usePickleCache=True, removePickleCache=True):
        self.assertTrue(DataTable.canReadExcel())
        xlsFile = os.path.join(os.path.dirname(__file__), 'Sample.xls')
        try:
            t = DataTable(xlsFile, usePickleCache=usePickleCache)
        finally:
            pickleFile = xlsFile + '.pickle.cache'
            self.assertIs(os.path.exists(pickleFile), usePickleCache)
            if usePickleCache and removePickleCache:
                os.remove(pickleFile)
        self.assertEqual(t[0][0], 1.0, t[0])



#!/usr/bin/env python3

import os
import sys


def compilePSP(*args):
    from .Context import PSPCLContext
    from .PSPCompiler import Compiler
    pspFilename = args[0]
    fil, ext = os.path.splitext(os.path.basename(pspFilename))
    classname = fil + '_' + ext
    pyFilename = classname + '.py'
    context = PSPCLContext(pspFilename)
    context.setClassName(classname)
    context.setPythonFileName(pyFilename)
    context.setPythonFileEncoding('utf-8')
    clc = Compiler(context)
    clc.compile()


if __name__ == '__main__':
    path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    sys.path.insert(0, path)
    from PSP.CompilePSP import compilePSP as main
    main(sys.argv[1])



import unittest

from WebUtils import Funcs


class TestFuncs(unittest.TestCase):

    def testHtmlEncode(self):
        f = Funcs.htmlEncode
        self.assertEqual(
            f('"1 < 2 & 2 > 1"'),
            '&quot;1 &lt; 2 &amp; 2 &gt; 1&quot;')
        self.assertEqual(
            f('quarks & co', (('&', 'and'),)),
            'quarks and co')
        self.assertEqual(f(None), '-')

        class Obj:
            @staticmethod
            def html():
                return 'bazong!'
        obj = Obj()
        self.assertEqual(f(obj), 'bazong!')
        obj.html = 'bazz!'
        self.assertEqual(f(obj), 'bazz!')

    def testHtmlEncodeStr(self):
        f = Funcs.htmlEncodeStr
        self.assertEqual(
            f('"1 < 2 & 2 > 1"'),
            '&quot;1 &lt; 2 &amp; 2 &gt; 1&quot;')
        self.assertEqual(
            f('quarks & co', (('&', 'and'),)),
            'quarks and co')

    def testHtmlDecode(self):
        f = Funcs.htmlDecode
        self.assertEqual(
            f('&quot;1 &lt; 2 &amp; 2 &gt; 1&quot;'),
            '"1 < 2 & 2 > 1"')
        self.assertEqual(
            f('quarks and co', (('&', 'and'),)),
            'quarks & co')

    def testHtmlRoundTrip(self):
        t = '<test>\n\t"1 < 2 & 2 > 1"\n</test>'
        self.assertEqual(Funcs.htmlDecode(Funcs.htmlEncode(t)), t)

    def testUrlEncode(self):
        f = Funcs.urlEncode
        self.assertEqual(f('"hello, world!"'), '%22hello%2C+world%21%22')

    def testUrlDecode(self):
        f = Funcs.urlDecode
        self.assertEqual(f('%22hello%2C+world%21%22'), '"hello, world!"')

    def testUrlRoundTrip(self):
        t = '<test>\n\t"50% = 50,50?"\n</test>'
        self.assertEqual(Funcs.urlDecode(Funcs.urlEncode(t)), t)

    def testHtmlDorDict(self):
        f = Funcs.htmlForDict
        self.assertEqual(
            f(dict(foo='bar', answer=42)),
            '<table class="NiceTable">\n'
            '<tr><th>answer</th><td>42</td></tr>\n'
            '<tr><th>foo</th><td>bar</td></tr>\n</table>')
        self.assertEqual(
            f(dict(foo='ba,zong', bar='ka;woom'),
              addSpace=dict(foo=',', bar=';')),
            '<table class="NiceTable">\n'
            '<tr><th>bar</th><td>ka; woom</td></tr>\n'
            '<tr><th>foo</th><td>ba, zong</td></tr>\n</table>')
        self.assertEqual(
            f(dict(foo='barbarabarbarabarbarabarbara'), maxValueLength=12),
            '<table class="NiceTable">\n'
            '<tr><th>foo</th><td>barbaraba...</td></tr>\n</table>')
        self.assertEqual(
            f(dict(foo='zing', bar='zang'),
              filterValueCallBack=lambda v, k, d: 'zung' if k == 'bar' else v),
            '<table class="NiceTable">\n'
            '<tr><th>bar</th><td>zung</td></tr>\n'
            '<tr><th>foo</th><td>zing</td></tr>\n</table>')
        self.assertEqual(
            f(dict(foo='bar'), topHeading='twinkle'),
            '<table class="NiceTable">\n'
            '<tr class="TopHeading"><th colspan="2">twinkle</th></tr>\n'
            '<tr><th>foo</th><td>bar</td></tr>\n</table>')
        self.assertEqual(
            f(dict(foo='bar'), topHeading=('key', 'value')),
            '<table class="NiceTable">\n'
            '<tr class="TopHeading"><th>key</th><th>value</th></tr>\n'
            '<tr><th>foo</th><td>bar</td></tr>\n</table>')
        self.assertEqual(
            f({'a & b': 'c & d'}),
            '<table class="NiceTable">\n'
            '<tr><th>a &amp; b</th><td>c &amp; d</td></tr>\n</table>')
        self.assertEqual(
            f({'a & b': 'c &amp; d'}, isEncoded=True),
            '<table class="NiceTable">\n'
            '<tr><th>a &amp; b</th><td>c &amp; d</td></tr>\n</table>')

    def testRequestURI(self):
        f = Funcs.requestURI
        self.assertEqual(
            f(dict(REQUEST_URI='http://w4py.org')), 'http://w4py.org')
        self.assertEqual(
            f(dict(SCRIPT_URL='http://w4py.org', QUERY_STRING='foo=bar')),
            'http://w4py.org?foo=bar')
        self.assertEqual(
            f(dict(SCRIPT_NAME='/test', QUERY_STRING='foo=bar')),
            '/test?foo=bar')

    def testNormURL(self):
        f = Funcs.normURL
        self.assertEqual(f('foo/bar'), 'foo/bar')
        self.assertEqual(f('/foo/bar'), '/foo/bar')
        self.assertEqual(f('foo/bar/'), 'foo/bar/')
        self.assertEqual(f('/foo/bar/../baz/../biz/./'), '/foo/biz/')
        self.assertEqual(f('/foo/bar/baz/biz/../../../'), '/foo/')
        self.assertEqual(f('/foo/./././././././'), '/foo/')
        self.assertEqual(f('/foo/../../../'), '/')
        self.assertEqual(f('foo///bar'), 'foo/bar')



"""The RoleUserManager mixin."""

from MiscUtils import NoDefault

from .RoleUser import RoleUser
from .Role import Role


class RoleUserManagerMixIn:
    """Mixin class for mapping names to roles.

    This mixin adds the functionality of keeping a dictionary mapping
    names to role instances. Several accessor methods are provided for this.
    """

    # region Init

    def __init__(self):
        self._roles = {}
        self.initUserClass()

    def initUserClass(self):
        """Invoked by __init__ to set the default user class to RoleUser."""
        self.setUserClass(RoleUser)

    # endregion Init

    # region Roles

    def addRole(self, role):
        if not isinstance(role, Role):
            raise TypeError(f'{role} is not a Role object')
        name = role.name()
        if name in self._roles:
            raise KeyError(f'role name {name!r} already exists')
        self._roles[name] = role

    def role(self, name, default=NoDefault):
        if default is NoDefault:
            return self._roles[name]
        return self._roles.get(name, default)

    def hasRole(self, name):
        return name in self._roles

    def delRole(self, name):
        del self._roles[name]

    def roles(self):
        return self._roles

    def clearRoles(self):
        self._roles = {}

    # endregion Roles



from Page import Page


class EmbeddedServlet(Page):
    """Test extra path info.

    This servlet serves as a test for "extra path info"-style URLs such as:

        http://localhost/Webware/Servlet/Extra/Path/Info

    Where the servlet is embedded in the URL, rather than being the last
    component. This servlet simply prints its fields.
    """

    def writeBody(self):
        fields = self.request().fields()
        self.writeln('<h2>EmbeddedServlet</h2>')
        self.writeln(f'<pre>{self.__class__.__doc__}</pre>')
        self.writeln(f'<h3>Fields: {len(fields)}</h3>')
        self.writeln('<ul>')
        enc = self.htmlEncode
        for key, value in fields.items():
            self.writeln(f'<li>{enc(key)} = {enc(value)}</li>')
        self.writeln('</ul>')



from .ExamplePage import ExamplePage

# Helper functions

gamma = 2.2  # an approximation for today's CRTs


def brightness(r, g, b):
    """Calculate brightness of RGB color."""
    r, g, b = [x / 255.0 for x in (r, g, b)]
    return (0.3*r**gamma + 0.6*g**gamma + 0.1*b**gamma)**(1/gamma)


def textColorForRgb(r, g, b):
    """Determine a good text font color for high contrast."""
    return 'white' if brightness(r, g, b) < 0.5 else 'black'


def rgbToHexColor(r, g, b):
    """Convert r, g, b to #RRGGBB."""
    return f'#{int(r):02X}{int(g):02X}{int(b):02X}'


def hexToRgbColor(h):
    """Convert #RRGGBB to r, g, b."""
    h = h.strip()
    if h.startswith('#'):
        h = h[1:]
    h = h[:2], h[2:4], h[4:]
    return [int(x, 16) for x in h]


def colorTable():
    """Prepare HTML for the color table"""
    numSteps = 6  # this gives the "web-safe" color palette
    steps = [255.0 * x / (numSteps - 1) for x in range(numSteps)]
    table = [
        '<p>Click on one of the colors below to set the background color.</p>',
        '<table style="margin-left:auto;margin-right:auto">']
    for r in steps:
        for g in steps:
            table.append('<tr>')
            for b in steps:
                bgColor = rgbToHexColor(r, g, b)
                textColor = textColorForRgb(r, g, b)
                table.append(
                    f'<td style="background-color:{bgColor};color:{textColor}"'
                    f' onclick="document.forms[0].bgColor.value=\'{bgColor}\';'
                    f'document.forms[0].submit()">{bgColor}</td>')
            table.append('</tr>')
    table.append('</table>')
    return '\n'.join(table)


class Colors(ExamplePage):
    """Colors demo.

    This class is a good example of caching. The color table that
    this servlet creates never changes, so the servlet caches this
    in the _colorTable class attribute. The original version of this
    example did no caching and was 12 times slower.
    """

    _colorTable = colorTable()

    def htBodyArgs(self):
        """Write the attributes of the body element.

        Overridden in order to throw in the custom background color
        that the user can specify in our form.
        """
        self._bgColor = self.request().field('bgColor', '#FFFFFF')
        try:
            r, g, b = hexToRgbColor(self._bgColor)
            self._textColor = textColorForRgb(r, g, b)
        except Exception:
            self._textColor = 'black'
        return f'style="color:black;background-color:{self._bgColor}"'

    def writeContent(self):
        """Write the actual content of the page."""
        self.write(f'''
            <div style="text-align:center;color:{self._textColor}">
            <h3>Color Table Demo</h3>
            <form action="Colors" method="post">
                <label for="bgColor">Background color:</label>
                <input type="text" name="bgColor" id="bgColor"
                       value="{self._bgColor}">
                <input type="submit" value="Go">
            </form>
            {self._colorTable}
            </div>
            ''')



import unittest

from Session import Session, SessionError

from .Transaction import Transaction


class TestSession(unittest.TestCase):

    def testInit(self):
        transaction = Transaction()
        session = Session(transaction)
        identifier = session.identifier()
        self.assertIsInstance(identifier, str)
        self.assertEqual(len(identifier), 47)
        session = Session(transaction, 'test-identifier')
        self.assertEqual(session.identifier(), 'test-identifier')
        with self.assertRaises(SessionError):
            Session(transaction, 'invalid+identifier')

    def testUniqueIds(self):
        transaction = Transaction()
        identifiers = set()
        for _count in range(3):
            session = Session(transaction)
            identifier = session.identifier()
            self.assertIsInstance(identifier, str)
            self.assertEqual(len(identifier), 47)
            parts = identifier.split('-')
            self.assertEqual(len(parts), 2)
            self.assertEqual(len(parts[0]), 14)
            self.assertEqual(len(parts[1]), 32)
            self.assertTrue(parts[0].isdigit())
            self.assertTrue(all(c in '0123456789abcdef' for c in parts[1]))
            self.assertNotIn(identifier, identifiers)
            identifiers.add(identifier)

    def testCreationAndAccessTime(self):
        transaction = Transaction()
        session = Session(transaction, 'test')
        t = session.creationTime()
        self.assertIsInstance(t, float)
        self.assertEqual(session.creationTime(), session.lastAccessTime())
        session = Session(transaction, 'test2')
        self.assertGreaterEqual(session.creationTime(), t)

    def testIsDirty(self):
        transaction = Transaction()
        session = Session(transaction, 'test')
        self.assertIs(session.isDirty(), False)
        session.setValue('test', None)
        self.assertIs(session.isDirty(), True)

    def testIsExpired(self):
        transaction = Transaction()
        session = Session(transaction, 'test')
        self.assertIs(session.isExpired(), False)
        session.expiring()
        self.assertIs(session.isExpired(), True)

    def testIsNew(self):
        transaction = Transaction()
        session = Session(transaction, 'test')
        self.assertIs(session.isNew(), True)
        session.awake(transaction)
        self.assertIs(session.isNew(), True)
        session.awake(transaction)
        self.assertIs(session.isNew(), False)

    def testTimeout(self):
        transaction = Transaction()
        session = Session(transaction, 'test')
        self.assertEqual(session.timeout(), 3600)
        session.setTimeout(3700)
        self.assertEqual(session.timeout(), 3700)
        session.invalidate()
        self.assertEqual(session.timeout(), 0)

    def testValues(self):
        transaction = Transaction()
        session = Session(transaction, 'test')
        self.assertRaises(KeyError, session.value, 'a')
        self.assertIsNone(session.value('a', None))
        self.assertEqual(session.value('a', 0), 0)
        self.assertEqual(session.value('a', 'foo'), 'foo')
        self.assertIs(session.hasValue('a'), False)
        session.setValue('a', None)
        self.assertIsNone(session.value('a'))
        self.assertIsNone(session.value('a', 'foo'))
        self.assertIs(session.hasValue('a'), True)
        session.delValue('a')
        self.assertRaises(KeyError, session.value, 'a')
        self.assertRaises(KeyError, session.delValue, 'a')
        self.assertIs(session.hasValue('a'), False)
        session.setValue('a', 'foo')
        self.assertEqual(session.value('a'), 'foo')
        self.assertEqual(session.value('a', None), 'foo')
        self.assertIs(session.hasValue('a'), True)
        session.setValue('b', 'bar')
        self.assertEqual(session.value('a'), 'foo')
        self.assertEqual(session.value('b'), 'bar')
        self.assertIs(session.hasValue('b'), True)
        self.assertIsInstance(session.values(), dict)
        self.assertEqual(session.values(), {'a': 'foo', 'b': 'bar'})

    def testDirectAccess(self):
        transaction = Transaction()
        session = Session(transaction, 'test')
        with self.assertRaises(KeyError):
            session['a']  # pylint: disable=pointless-statement
        self.assertNotIn('a', session)
        session['a'] = 'foo'
        self.assertEqual(session['a'], 'foo')
        self.assertIn('a', session)
        self.assertEqual(session.value('a'), 'foo')
        self.assertTrue(session.hasValue('a'))
        del session['a']
        with self.assertRaises(KeyError):
            session['a']  # pylint: disable=pointless-statement
        self.assertNotIn('a', session)
        self.assertFalse(session.hasValue('a'))

    def testSessionEncode(self):
        transaction = Transaction()
        session = Session(transaction, 'test')
        self.assertEqual(session.sessionEncode('myurl'), 'myurl?_SID_=test')
        self.assertEqual(
            session.sessionEncode('http://localhost/Webware/Servlet?a=1&b=2'),
            'http://localhost/Webware/Servlet?a=1&b=2&_SID_=test')



"""A bunch of utility functions for the PSP generator.

Copyright (c) by Jay Love, 2000 (mailto:jsliv@jslove.org)

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee or royalty is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation or portions thereof, including modifications,
that you make.

This software is based in part on work done by the Jakarta group.
"""


class PSPParserException(Exception):
    """PSP parser error."""


def removeQuotes(s):
    return s.replace(r'%\\>', '%>')


def isExpression(s):
    """Check whether this is a PSP expression."""
    return s.startswith('<%=') and s.endswith('%>')


def getExpr(s):
    """Get the content of a PSP expression."""
    if s.startswith('<%=') and s.endswith('%>'):
        return s[3:-2]
    return ''


def checkAttributes(tagType, attrs, validAttrs):
    """Check for mandatory and optional atributes."""
    attrs = set(attrs)
    mandatoryAttrs = validAttrs[0]
    for attr in mandatoryAttrs:
        try:
            attrs.remove(attr)
        except KeyError:
            raise PSPParserException(
                f'{tagType}: Mandatory attribute {attr} missing') from None
    optionalAttrs = validAttrs[1]
    for attr in attrs:
        if attr not in optionalAttrs:
            raise PSPParserException('{tagType}: Invalid attribute {attr}')


def splitLines(text, keepends=False):
    """Split text into lines."""
    return text.splitlines(keepends)


def startsNewBlock(line):
    """Determine whether a code line starts a new block."""
    line = line.strip()
    if line.startswith('#'):
        return False
    try:
        compile(line, '<string>', 'exec')
    except SyntaxError:
        try:
            compile(line + '\n    pass', '<string>', 'exec')
        except Exception:
            pass
        else:
            return True
    else:
        return False
    return line.endswith(':')


def normalizeIndentation(pySource):
    """Take code block that may be too indented and move it all to the left."""
    lines = splitLines(pySource, True)
    minIndent = None
    indents = []
    for line in lines:
        strippedLine = line.lstrip(' \t')
        indent = len(line) - len(strippedLine)
        indents.append(len(line) - len(strippedLine))
        if strippedLine.lstrip('\r\n') and not strippedLine.startswith('#'):
            if minIndent is None or indent < minIndent:
                minIndent = indent
                if not minIndent:
                    break
    if minIndent:
        pySource = ''.join(
            line[min(minIndent, indent):]
            for line, indent in zip(lines, indents))
    return pySource



from types import FunctionType, MethodType


def MixIn(pyClass, mixInClass, makeAncestor=False, mixInSuperMethods=False):
    """Mixes in the attributes of the mixInClass into the pyClass.

    These attributes are typically methods (but don't have to be).
    Note that private attributes, denoted by a double underscore,
    are not mixed in. Collisions are resolved by the mixInClass'
    attribute overwriting the pyClass'. This gives mix-ins the power
    to override the behavior of the pyClass.

    After using MixIn(), instances of the pyClass will respond to
    the messages of the mixInClass.

    An assertion fails if you try to mix in a class with itself.

    The pyClass will be given a new attribute mixInsForCLASSNAME
    which is a list of all mixInClass' that have ever been installed,
    in the order they were installed. You may find this useful
    for inspection and debugging.

    You are advised to install your mix-ins at the start up
    of your program, prior to the creation of any objects.
    This approach will result in less headaches. But like most things
    in Python, you're free to do whatever you're willing to live with. :-)

    There is a bitchin' article in the Linux Journal, April 2001,
    "Using Mix-ins with Python" by Chuck Esterbrook,
    which gives a thorough treatment of this topic.

    An example, that resides in the Webware MiddleKit plug-in, is
    MiddleKit.Core.ModelUser.py, which install mix-ins for SQL interfaces.
    Search for "MixIn(".

    If makeAncestor is True, then a different technique is employed:
    a new class is created and returned that is the same as the given pyClass,
    but will have the mixInClass added as its first base class.
    Note that this is different from the behavior in legacy Webware
    versions, where the __bases__ attribute of the pyClass was changed.
    You probably don't need to use this and if you do, be aware that your
    mix-in can no longer override attributes/methods in pyClass.

    If mixInSuperMethods is True, then support will be enabled for you to
    be able to call the original or "parent" method from the mixed-in method.
    This is done like so::

        class MyMixInClass:
        def foo(self):
            MyMixInClass.mixInSuperFoo(self)  # call the original method
            # now do whatever you want
    """
    if mixInClass is pyClass:
        raise TypeError('mixInClass is the same as pyClass')
    if makeAncestor:
        if mixInClass not in pyClass.__bases__:
            return type(pyClass.__name__,
                        (mixInClass,) + pyClass.__bases__,
                        dict(pyClass.__dict__))
    else:
        # Recursively traverse the mix-in ancestor classes in order
        # to support inheritance
        for baseClass in reversed(mixInClass.__bases__):
            if baseClass is not object:
                MixIn(pyClass, baseClass)

        # Track the mix-ins made for a particular class
        attrName = 'mixInsFor' + pyClass.__name__
        mixIns = getattr(pyClass, attrName, None)
        if mixIns is None:
            mixIns = []
            setattr(pyClass, attrName, mixIns)

        # Record our deed for future inspection
        mixIns.append(mixInClass)

        # Install the mix-in methods into the class
        for name in dir(mixInClass):
            # skip private members, but not __repr__ et al:
            if name.startswith('__'):
                if not name.endswith('__'):
                    continue  # private
                member = getattr(mixInClass, name)
                if not isinstance(member, MethodType):
                    continue  # built in or descriptor
            else:
                member = getattr(mixInClass, name)
            if isinstance(member, (FunctionType, MethodType)):
                if mixInSuperMethods:
                    if hasattr(pyClass, name):
                        origMember = getattr(pyClass, name)
                        setattr(mixInClass, 'mixInSuper' +
                                name[0].upper() + name[1:], origMember)
            setattr(pyClass, name, member)



import os

from ConfigurableForServerSidePath import ConfigurableForServerSidePath


class MockImportManager:

    def recordFile(self, filename, isfile=None):
        pass


defaultConfig = dict(
    CacheDir='Cache',
    PlugIns=['MiscUtils', 'WebUtils', 'TaskKit', 'UserKit', 'PSP'],
    PrintPlugIns=False
)


class MockApplication(ConfigurableForServerSidePath):
    """
    A minimal implementation which is compatible with Application
    and which is sufficient to load plugins.
    """

    def __init__(self, path=None, settings=None, development=None):
        ConfigurableForServerSidePath.__init__(self)
        if path is None:
            path = os.getcwd()
        self._serverSidePath = os.path.abspath(path)
        self._webwarePath = os.path.abspath(os.path.dirname(__file__))
        if development is None:
            development = bool(os.environ.get('WEBWARE_DEVELOPMENT'))
        self._development = development
        appConfig = self.config()
        if settings:
            appConfig.update(settings)
        self._cacheDir = self.serverSidePath(
            self.setting('CacheDir') or 'Cache')
        from MiscUtils.PropertiesObject import PropertiesObject
        props = PropertiesObject(os.path.join(
            self._webwarePath, 'Properties.py'))
        self._webwareVersion = props['version']
        self._webwareVersionString = props['versionString']
        self._imp = MockImportManager()
        for path in (self._cacheDir,):
            if path and not os.path.exists(path):
                os.makedirs(path)

    def defaultConfig(self):
        return defaultConfig

    def configReplacementValues(self):
        """Get config values that need to be escaped."""
        return dict(
            ServerSidePath=self._serverSidePath,
            WebwarePath=self._webwarePath,
            Development=self._development)

    def configFilename(self):
        return self.serverSidePath('Configs/Application.config')

    def serverSidePath(self, path=None):
        if path:
            return os.path.normpath(
                os.path.join(self._serverSidePath, path))
        return self._serverSidePath

    def hasContext(self, _name):
        return False

    def addServletFactory(self, factory):
        pass



import os

from .PSPExamplePage import PSPExamplePage


class View(PSPExamplePage):
    """View the source of a PSP page.

    For each PSP example, you will see a sidebar with various menu items,
    one of which is "View source of <em>example</em>". This link points to the
    View servlet and passes the filename of the current servlet. The View
    servlet then loads that PSP file's source code and displays it in the
    browser for your viewing pleasure.

    Note that if the View servlet isn't passed a PSP filename, it prints the
    View's docstring which you are reading right now.
    """

    def writeContent(self):
        req = self.request()
        if req.hasField('filename'):
            filename = req.field('filename')
            basename = os.path.basename(filename)
            filename = self.request().serverSidePath(basename)
            if not os.path.exists(filename):
                self.write('<p style="color:red">'
                           f'No such file {basename} exists</p>')
                return
            text = open(filename).read()
            text = self.htmlEncode(text)
            text = text.replace('\n', '<br>').replace('\t', ' '*4)
            self.write(f'<pre>{text}</pre>')
        else:
            doc = self.__class__.__doc__.split('\n', 1)
            doc[1] = '</p>\n<p>'.join(doc[1].split('\n\n'))
            self.writeln(f'<h2>{doc[0]}</h2>\n<p>{doc[1]}</p>')



#!/usr/bin/env python3

"""The Webware command line interface"""

import argparse

from .MakeAppWorkDir import addArguments as addMakeArguments, make
from .WaitressServer import addArguments as addServeArguments, serve
from ..Properties import version as versionTuple
from ..MiscUtils.PropertiesObject import versionString


def main(args=None):
    """Evaluate the command line arguments and execute subcommand."""
    version = versionString(versionTuple)
    parser = argparse.ArgumentParser(description="Webware CLI")
    parser.add_argument('-v', '--version', action='version',
                        version=f'Webware for Python {version}')
    subparsers = parser.add_subparsers(
        dest='command', title="Webware subcommands",
        help="name of the subcommand")
    serveParser = subparsers.add_parser(
        'serve', help="Serve a Webware application using waitress")
    addServeArguments(serveParser)
    makeParser = subparsers.add_parser(
        'make', help="Make a Webware application working directory")
    addMakeArguments(makeParser)
    args = parser.parse_args(args)
    command = args.command
    del args.command
    if command == 'make':
        make(args)
    elif command == 'serve':
        serve(args)


if __name__ == '__main__':
    main()



import time
import http.client

from SidebarPage import SidebarPage


class TestIMS(SidebarPage):

    def cornerTitle(self):
        return 'Testing'

    def error(self, msg):
        self.write(f'<p style="color:red">{self.htmlEncode(msg)}</p>')

    def writeMsg(self, msg):
        self.write(f'<p>{self.htmlEncode(msg)}</p>')

    def writeTest(self, msg):
        self.write(f'<h4>{msg}</h4>')

    def getDoc(self, path, headers=None):
        if headers is None:
            headers = {}
        con = self._httpConnection(self._host)
        con.request('GET', path, headers=headers)
        return con.getresponse()

    def writeContent(self):
        self.writeln('<h2>Test If-Modified-Since support in Webware</h2>')
        if self.request().environ().get('paste.testing'):
            self.writeln('<p>This test requires a running web server.</p/>')
            return
        d = self.request().serverDictionary()
        self._host = d['HTTP_HOST']  # includes the port
        self._httpConnection = (
            http.client.HTTPSConnection
            if d.get('HTTPS', '').lower() == 'on'
            else http.client.HTTPConnection)
        servletPath = self.request().servletPath()
        # pick a static file which is served up by Webware's UnknownFileHandler
        self.runTest(f'{servletPath}/PSP/Examples/psplogo.png')

    def runTest(self, path):
        self.writeTest(f'Opening <code>{path}</code>')
        rsp = self.getDoc(path)
        originalSize = size = len(rsp.read())
        if rsp.status != 200:
            self.error(f'Expected status of 200, received {rsp.status}.')
            return
        if size > 0:
            self.writeMsg(
                f'Received: {rsp.status} rsp.reason{rsp.reason},'
                f' document size = {size} (as expected).')
        else:
            self.error(f'Document size is: {size}')
            return
        lastMod = rsp.getheader('Last-Modified', '')
        if lastMod:
            self.writeMsg(f'Last modified: {lastMod}')
        else:
            self.error('No Last-Modified header found.')
            return
        # Retrieve document again with IMS and expect a 304 not modified
        self.writeTest(
            f'Opening <code>{path}</code><br>'
            f'with If-Modified-Since: {lastMod}')
        rsp = self.getDoc(path, {'If-Modified-Since': lastMod})
        size = len(rsp.read())
        if rsp.status != 304:
            self.error(f'Expected status of 304, received {rsp.status}.')
            return
        if size:
            self.error(f'Expected 0 length document, received {size} bytes.')
            return
        self.writeMsg(
            f'Received {rsp.status} {rsp.reason},'
            f' document size = {size} (as expected).')
        arpaFormat = '%a, %d %b %Y %H:%M:%S GMT'
        t = time.strptime(lastMod, arpaFormat)
        t = (t[0] - 1,) + t[1:]  # one year before last modification
        beforeMod = time.strftime(arpaFormat, time.gmtime(time.mktime(t)))
        self.writeTest(
            f'Opening <code>{path}</code><br>'
            f'with If-Modified-Since: {beforeMod}')
        rsp = self.getDoc(path, {'If-Modified-Since': beforeMod})
        size = len(rsp.read())
        lastMod = rsp.getheader('Last-Modified', '')
        self.writeMsg(f'Last modified: {lastMod}')
        if rsp.status != 200:
            self.error(
                f'Expected status of 200, received {rsp.status} {rsp.reason}.')
            return
        if size != originalSize:
            self.error(
                f'Received: {rsp.status} {rsp.reason},'
                f' document size = {size}, expected size = {originalSize}.')
            return
        self.writeMsg(
            f'Received: {rsp.status} {rsp.reason},'
            f' document size = {size} (as expected).')
        self.writeTest(f'{self.__class__.__name__} passed.')



name = 'Python Server Pages'

status = 'stable'

synopsis = (
    "A Python Server Page (or PSP) is an HTML document"
    " with interspersed Python instructions that are interpreted"
    " as a template to generate dynamic content."
    " PSP is analogous to PHP, Microsoft's ASP and Sun's JSP."
    " PSP sits on top of (and requires) Webware and therefore"
    " benefits from its features.")

webwareConfig = {
    'examplePages': [
        'Hello',
        'Braces',
        'PSPTests',
        'PSPTests-Braces',
    ]
}



from SidebarPage import SidebarPage


class Main(SidebarPage):
    """Testing - read TestCases.data and display them.

    TO DO
      * Reload TestCases.data only load when modified (by checking mod date).
    """

    def title(self):
        return 'Testing'

    def cornerTitle(self):
        return 'Webware Testing'

    def writeContent(self):
        self.writeln('<h2 style="text-align:center">Test cases</h2>')
        self.writeTestCases()
        self.writeNotes()

    def writeTestCases(self):
        wr = self.writeln
        req = self.request()
        servletPath = req.servletPath()
        filename = self.serverSidePath('TestCases.data')
        self._cases = self.readFileNamed(filename)
        wr('<table style="margin-left:auto;margin-right:auto">')
        wr('<tr><th>#</th><th>URL</th><th>Expectation</th></tr>')
        caseNum = 1
        for case in self._cases:
            # For each URL, fix it up and make a name. Put in urls list.
            urls = []
            for url in case['URLs']:
                url = servletPath + url
                if url:
                    urlName = self.htmlEncode(url)
                    urls.append((url, urlName))
            if not urls:
                continue
            expectation = case['Expectation']
            bgColor = ['eee', 'ddd'][caseNum % 2]
            wr(f'<tr style="background-color:#{bgColor}">'
               f'<td>{caseNum}.</td><td>')
            wr('<br>'.join(f'<a href="{url}">{urlName}</a>'
                           for url, urlName in urls))
            wr(f'</td><td>{expectation}</td></tr>')
            caseNum += 1
        wr('</table>')

    def readFileNamed(self, filename):
        """Return a list of test cases.

        Each of them is a dictionary, as defined the given file.
        See TestCases.data for information on the format.
        """
        with open(filename) as f:
            cases = self.readFile(f)
        return cases

    def readFile(self, file):
        return self.readContent(file.read())

    def readContent(self, content):
        lines = [line.strip() for line in content.splitlines()]
        lineNum = 1
        cases = []
        urls = []
        for line in lines:
            if line and line[0] != '#':
                if line[-1] == '\\':  # continuation line;
                    # means there are multiple URLs for this case
                    urls.append(line[:-1].strip())
                else:
                    parts = line.split('-->')
                    if len(parts) != 2:
                        self.error(f'Invalid line at {lineNum}.')
                    urls.append(parts[0].strip())
                    cases.append(dict(
                        URLs=urls, Expectation=parts[1].strip()))
                    urls = []  # reset list of URLs
            lineNum += 1
        return cases

    def writeNotes(self):
        self.writeln('''<h4>Notes</h4>
<ul>
<li>Test all links in all pages of all contexts (Examples, Admin, Testing...),
including links found in the headers and footers of the pages.</li>
<li>Test functionality of interactive pages, like CountVisits and ListBox.</li>
<li>Test each link more than once.</li>
</ul>''')

    def error(self, msg):
        raise Exception(msg)



from .ExamplePage import ExamplePage

debug = False


class ListBox(ExamplePage):
    """List box example.

    This page provides a list box interface with controls
    for changing its size and adding and removing items.

    The source is a good example of how to use awake() and actions.

    It also shows how to avoid repeated execution on refresh/reload.
    """

    def awake(self, transaction):
        ExamplePage.awake(self, transaction)
        session = self.session()
        if session.hasValue('vars'):
            self._vars = session.value('vars')
        else:
            self._vars = dict(
                items=[], height=10, width=250,
                newCount=1, formCount=1)
            session.setValue('vars', self._vars)
        self._error = None

    def writeContent(self):
        enc, wr = self.htmlEncode, self.writeln
        wr('<div style="text-align:center">')
        if debug:
            wr('<p>fields = {}</p>'.format(enc(str(self.request().fields()))))
            wr('<p>vars = {}</p>'.format(enc(str(self._vars))))
        # Intro text is provided by our class' doc string:
        intro = self.__class__.__doc__.strip().split('\n\n')
        wr('<h2>{}</h2>'.format(intro.pop(0)))
        for s in intro:
            wr('<p>{}</p>'.format('<br>'.join(
                s.strip() for s in s.splitlines())))
        wr('<p style="color:red">{}</p>'.format(self._error or '&nbsp;'))
        wr('''
<form action="ListBox" method="post">
<input name="formCount" type="hidden" value="{formCount}">
<select multiple name="items" size="{height}"
style="width:{width}pt;text-align:center">
'''.format(**self._vars))
        index = 0
        for item in self._vars['items']:
            wr('<option value="{}">{}</option>'.format(
                index, enc(item['name'])))
            index += 1
        if not index:
            wr('<option value="" disabled>--- empty ---</option>')
        wr('''
</select>
<p>
<input name="_action_new" type="submit" value="New">
<input name="_action_delete" type="submit" value="Delete">
&nbsp; &nbsp; &nbsp;
<input name="_action_taller" type="submit" value="Taller">
<input name="_action_shorter" type="submit" value="Shorter">
&nbsp; &nbsp; &nbsp;
<input name="_action_wider" type="submit" value="Wider">
<input name="_action_narrower" type="submit" value="Narrower">
</p>
</form>
</div>
''')

    @staticmethod
    def heightChange():
        return 1

    @staticmethod
    def widthChange():
        return 30

    # Commands

    def new(self):
        """Add a new item to the list box."""
        self._vars['items'].append(dict(
            name='New item {}'.format(self._vars['newCount'])))
        self._vars['newCount'] += 1
        self.writeBody()

    def delete(self):
        """Delete the selected items in the list box."""
        req = self.request()
        if req.hasField('items'):
            indices = req.field('items')
            if not isinstance(indices, list):
                indices = [indices]
            try:
                indices = list(map(int, indices))  # convert strings to ints
            except ValueError:
                indices = []
            # remove the objects:
            for index in sorted(indices, reverse=True):
                del self._vars['items'][index]
        else:
            self._error = 'You must select a row to delete.'
        self.writeBody()

    def taller(self):
        self._vars['height'] += self.heightChange()
        self.writeBody()

    def shorter(self):
        if self._vars['height'] > 2:
            self._vars['height'] -= self.heightChange()
        self.writeBody()

    def wider(self):
        self._vars['width'] += self.widthChange()
        self.writeBody()

    def narrower(self):
        if self._vars['width'] >= 60:
            self._vars['width'] -= self.widthChange()
        self.writeBody()

    # Actions

    def actions(self):
        acts = ExamplePage.actions(self)
        # check whether form is valid (no repeated execution)
        try:
            formCount = int(self.request().field('formCount'))
        except (KeyError, ValueError):
            formCount = 0
        if formCount == self._vars['formCount']:
            acts.extend([
                'new', 'delete', 'taller', 'shorter', 'wider', 'narrower'])
            self._vars['formCount'] += 1
        return acts



"""Universal error class."""


class Error(dict):
    """Universal error class.

    An error is a dictionary-like object, containing a specific
    user-readable error message and an object associated with it.
    Since Error inherits dict, other informative values can be arbitrarily
    attached to errors. For this reason, subclassing Error is rare.

    Example::

        err = Error(user, 'Invalid password.')
        err['time'] = time.time()
        err['attempts'] = attempts

    The object and message can be accessed via methods::

        print(err.object())
        print(err.message())

    When creating errors, you can pass None for both object and message.
    You can also pass additional values, which are then included in the error::

        >>> err = Error(None, 'Too bad.', timestamp=time.time())
        >>> err.keys()
        ['timestamp']

    Or include the values as a dictionary, instead of keyword arguments::

        >>> info = {'timestamp': time.time()}
        >>> err = Error(None, 'Too bad.', info)

    Or you could even do both if you needed to.
    """

    def __init__(self, obj, message, valueDict=None, **valueArgs):
        """Initialize the error.

        Takes the object the error occurred for, and the user-readable
        error message. The message should be self sufficient such that
        if printed by itself, the user would understand it.
        """
        dict.__init__(self)
        self._object = obj
        self._message = message
        if valueDict:
            self.update(valueDict)
        self.update(valueArgs)

    def object(self):
        """Get the object the error occurred for."""
        return self._object

    def message(self):
        """Get the user-readable error message."""
        return self._message

    def __repr__(self):
        return (f'ERROR(object={self._object!r};'
                f' message={self._message!r}; data={dict(self)!r})')

    def __str__(self):
        return f'ERROR: {self._message}'

    def __bool__(self):
        return True



"""Session store using the Memcached memory object caching system."""

from warnings import warn
from pickle import HIGHEST_PROTOCOL as maxPickleProtocol

try:
    import memcache  # pylint: disable=import-error
except Exception as e:
    raise ImportError(
        "For using Memcached sessions,"
        " python-memcached must be installed.") from e

from MiscUtils import NoDefault

from SessionStore import SessionStore

debug = False


class SessionMemcachedStore(SessionStore):
    """A session store using Memcached.

    Stores the sessions in a single Memcached store using 'last write wins'
    semantics. This increases fault tolerance and allows server clustering.
    In clustering configurations with concurrent writes for the same
    session(s) the last writer will always overwrite the session.

    The keys are prefixed with a configurable namespace, allowing you to
    store other data in the same Memcached system.

    Cleaning/timing out of sessions is performed by Memcached itself since
    no single application can know about the existence of all sessions or
    the last access for a given session. Besides it is built in Memcached
    functionality. Consequently, correct sizing of Memcached is necessary
    to hold all user's session data.

    Due to the way Memcached works, methods requiring access to the keys
    or for clearing the store do not work. You can configure whether you
    want to ignore such calls or raise an error in this case. By default,
    you will get a warning. It would be possible to emulate these functions
    by storing additional data in the memcache, such as a namespace counter
    or the number or even the full list of keys. However, if you are using
    more than one application instance, this would require fetching that data
    every time, since we cannot know whether another instance changed it.
    So we refrained from doing such sophisticated trickery and instead kept
    the implementation intentionally very simple and fast.

    You need to install python-memcached to be able to use this module:
    https://www.tummy.com/software/python-memcached/
    You also need a Memcached server: https://memcached.org/

    Contributed by Steve Schwarz, March 2010.
    Small improvements by Christoph Zwerschke, April 2010.
    """

    # region Init

    def __init__(self, app):
        SessionStore.__init__(self, app)

        # the list of memcached servers
        self._servers = app.setting('MemcachedServers', ['localhost:11211'])

        # timeout in seconds
        self._sessionTimeout = app.setting('SessionTimeout', 180) * 60

        # the memcached "namespace" used by our store
        # you can add an integer counter for expiration
        self._namespace = app.setting(
            'MemcachedNamespace', 'WebwareSession:') or ''

        # when trying to iterate over the Memcached store,
        # you can trigger an error or a warning
        self._onIteration = app.setting('MemcachedOnIteration', 'Warning')

        self._client = memcache.Client(
            self._servers, debug=debug, pickleProtocol=maxPickleProtocol)

    # endregion Init

    # region Access

    def __len__(self):
        """Return the number of sessions in the store.

        Not supported by Memcached (see FAQ for explanation).
        """
        if debug:
            print(">> len()")
        return len(self.keys())

    def __getitem__(self, key):
        """Get a session item, reading it from the store."""
        if debug:
            print(f">> getitem({key})")
        # returns None if key non-existent or no server to contact
        try:
            value = self._client.get(self.mcKey(key))
        except Exception:
            value = None
        if value is None:
            # SessionStore expects KeyError when no result
            raise KeyError(key)
        return value

    def __setitem__(self, key, value):
        """Set a session item, writing it to the store."""
        if debug:
            print(f">> setitem({key}, {value})")
        dirty = value.isDirty()
        if self._alwaysSave or dirty:
            if dirty:
                value.setDirty(False)
            try:
                if not self._client.set(
                        self.mcKey(key), value, time=self._sessionTimeout):
                    raise ValueError("Setting value in the memcache failed.")
            except Exception as exc:
                if dirty:
                    value.setDirty()
                # Not able to store the session is a failure
                print(f"Error saving session {key!r} to memcache: {exc}")
                self.application().handleException()

    def __delitem__(self, key):
        """Delete a session item from the store.

        Note that in contracts with SessionFileStore,
        not finding a key to delete isn't a KeyError.
        """
        if debug:
            print(f">> delitem({key})")
        session = self[key]
        if not session.isExpired():
            session.expiring()
        try:
            if not self._client.delete(self.mcKey(key)):
                raise ValueError("Deleting value from the memcache failed.")
        except Exception as exc:
            # Not able to delete the session is a failure
            print(f"Error deleting session {key!r} from memcache: {exc}")
            self.application().handleException()

    def __contains__(self, key):
        """Check whether the session store has a given key."""
        if debug:
            print(f">> contains({key})")
        try:
            return self._client.get(self.mcKey(key)) is not None
        except Exception:
            return False

    def __iter__(self):
        """Return an iterator over the stored session keys.

        Not supported by Memcached (see FAQ for explanation).
        """
        if debug:
            print(">> iter()")
        onIteration = self._onIteration
        if onIteration:
            err = 'Memcached does not support iteration.'
            if onIteration == 'Error':
                raise NotImplementedError(err)
            warn(err)
        return iter([])

    def keys(self):
        """Return a list with the keys of all the stored sessions.

        Not supported by Memcached (see FAQ for explanation).
        """
        if debug:
            print(">> keys()")
        return list(iter(self))

    def clear(self):
        """Clear the session store, removing all of its items.

        Not supported by Memcached. We could emulate this by incrementing
        an additional namespace counter, but then we would need to fetch
        the current counter from the memcache before every access in order
        to keep different application instances in sync.
        """
        if debug:
            print(">> clear()")
        if self._onIteration:
            err = 'Memcached does not support clearing the store.'
            if self._onIteration == 'Error':
                raise NotImplementedError(err)
            warn(err)

    def setdefault(self, key, default=None):
        """Return value if key available, else default (also setting it)."""
        if debug:
            print(f">> setdefault({key}, {default})")
        try:
            return self[key]
        except KeyError:
            self[key] = default
            return default

    def pop(self, key, default=NoDefault):
        """Return value if key available, else default (also remove key)."""
        if debug:
            print(f">> pop({key}, {default})")
        if default is NoDefault:
            value = self[key]
            del self[key]
            return value
        try:
            value = self[key]
        except KeyError:
            return default
        else:
            del self[key]
            return value

    # endregion Access

    # region Application support

    def storeSession(self, session):
        """Save potentially changed session in the store."""
        if debug:
            print(f">> storeSession({session})")
        self[session.identifier()] = session

    def storeAllSessions(self):
        """Permanently save all sessions in the store.

        Should be used (only) when the application server is shut down.
        This closes the connection to the Memcached servers.
        """
        if debug:
            print(">> storeAllSessions()")
        self._client.disconnect_all()

    def cleanStaleSessions(self, _task=None):
        """Clean stale sessions.

        Memcached does this on its own, so we do nothing here.
        """
        if debug:
            print(">> cleanStaleSessions()")

    # endregion Application support

    # region Auxiliary methods

    def mcKey(self, key):
        """Create the real key with namespace to be used with Memcached."""
        return self._namespace + key

    # endregion Auxiliary methods



"""The RoleUserManagerToFile class."""

from .RoleUserManagerMixIn import RoleUserManagerMixIn
from .UserManagerToFile import UserManagerToFile


class RoleUserManagerToFile(RoleUserManagerMixIn, UserManagerToFile):
    """See the base classes for more information."""

    def __init__(self, userClass=None):
        UserManagerToFile.__init__(self, userClass)
        RoleUserManagerMixIn.__init__(self)



"""Non-automatic test-run for the TaskKit scheduler."""

from time import time, localtime, sleep

from TaskKit.Scheduler import Scheduler
from TaskKit.Task import Task


class SimpleTask(Task):

    increasingPeriod = False

    def run(self):
        if self.proceed():
            print(self.name(), time())
            if self.increasingPeriod:
                self.handle().setPeriod(self.handle().period() + 2)
        else:
            print("Should not proceed", self.name())
            print(f"proceed for {self.name()}={self.proceed()}",
                  f"isRunning={self._handle._isRunning}")


class LongTask(Task):

    def run(self):
        while True:
            sleep(2)
            print(f"proceed for {self.name()}={self.proceed()}",
                  f"isRunning={self._handle._isRunning}")
            if self.proceed():
                print(">>", self.name(), time())
            else:
                print("Should not proceed:", self.name())
                return


def main():
    scheduler = Scheduler()
    scheduler.start()
    scheduler.addPeriodicAction(time(), 1, SimpleTask(), 'SimpleTask1')
    scheduler.addTimedAction(time() + 3, SimpleTask(), 'SimpleTask2')
    scheduler.addActionOnDemand(LongTask(), 'LongTask')
    sleep(4)
    print("Demanding 'LongTask'")
    scheduler.runTaskNow('LongTask')
    sleep(1)
    print("Stopping 'LongTask'")
    scheduler.stopTask('LongTask')
    sleep(2)
    print("Deleting 'SimpleTask1'")
    scheduler.unregisterTask('SimpleTask1')
    sleep(2)
    print("Waiting one minute for 'DailyTask'")
    scheduler.addDailyAction(
        localtime()[3], localtime()[4] + 1, SimpleTask(), "DailyTask")
    sleep(62)
    print("Calling stop")
    scheduler.stop()
    sleep(2)
    print("Test Complete")


if __name__ == '__main__':
    main()



from .ExamplePage import ExamplePage


class Welcome(ExamplePage):

    def writeContent(self):
        wr = self.writeln
        version = self.application().webwareVersionString()
        wr(f'<h2>Welcome to Webware for Python {version}!</h2>')
        path = self.request().servletPath()
        wr(f'''
        <p>Along the side of this page you will see various links
        that will take you to:</p>
        <ul>
            <li>The different Webware examples.</li>
            <li>The source code of the current example.</li>
            <li>Whatever contexts have been configured.
                Each context represents a distinct set of web pages,
                usually given a descriptive name.</li>
            <li>External sites, such as the Webware home page.</li>
        </ul>
        <p>The <a href="{path}/Admin/">Admin</a> context is particularly
        interesting because it takes you to the administrative pages
        for the Webware application where you can review logs,
        configuration, plug-ins, etc.</p>
        ''')
        req = self.request()
        extraURLPath = req.extraURLPath()
        if extraURLPath and extraURLPath != '/':
            wr('''
            <p><strong>Note:</strong>
            extraURLPath information was found on the URL,
            and a servlet was not found to process it.
            Processing has been delegated to this servlet.</p>
            ''')
            wr('<ul>')
            wr(f'<li>serverSidePath: <code>{req.serverSidePath()}</code></li>')
            wr(f'<li>extraURLPath: <code>{extraURLPath}</code></li>')
            wr('</ul>')



import os

from SidebarPage import SidebarPage


class AdminPage(SidebarPage):
    """AdminPage

    This is the abstract superclass of all Webware administration pages.

    Subclasses typically override title() and writeContent(), but may
    customize other methods.
    """

    def cornerTitle(self):
        return 'Webware Admin'

    def writeSidebar(self):
        self.writeAdminMenu()
        self.writeWebwareSidebarSections()

    def writeAdminMenu(self):
        self.menuHeading('Admin')
        self.menuItem('Home', 'Main')
        self.menuItem(
            'Activity log', 'Access', self.fileSize('ActivityLogFilename'))
        self.menuItem(
            'Error log', 'Errors', self.fileSize('ErrorLogFilename'))
        self.menuItem('Config', 'Config')
        self.menuItem('Plug-ins', 'PlugIns')
        self.menuItem('Servlet Cache', 'ServletCache')
        self.menuItem('Application Control', 'AppControl')
        self.menuItem('Logout', 'Main?logout=yes')

    def fileSize(self, filename):
        """Utility method for writeMenu() to get the size of a config file.

        Returns an HTML string.
        """
        filename = self.application().setting(filename)
        if os.path.exists(filename):
            size = '{:0.0f} KB'.format(os.path.getsize(filename) / 1024)
        else:
            size = 'not existent'
        return f'<span style="font-size:smaller">({size})</span>'

    def loginDisabled(self):
        """Return None if login is enabled, else a message about why not."""
        if self.application().setting('AdminPassword'):
            return None
        return (
            '<p>Logins to admin pages are disabled until'
            ' you supply an AdminPassword in Application.config.</p>')



from MiscUtils.Funcs import asclocaltime

from .ExamplePage import ExamplePage


class ShowTime(ExamplePage):

    def writeContent(self):
        self.write('<h4>The current time is:</h4>')
        self.write('<h5 style="color:green">', asclocaltime(), '</h5>')
